import torch
from .ops import op, op_dgl
import torch.nn.functional as F
from torch.nn import Identity

class ConvDGL(torch.nn.Module):
    def __init__(self, space, input_dim, output_dim, pos=0, act=Identity(), dropout=0.0, bn=False, res=False, track=False, edge_feat=16, **kwargs) -> None:
        super().__init__()
        self.pos = pos
        self.act = act
        self.bn = bn
        if dropout > 0:
            self.dropout = torch.nn.Dropout(p=dropout)
        else:
            self.dropout = Identity()
        self.space = space
        self.res = res

        # span a space
        self.core = torch.nn.ModuleDict({name: op_dgl(name, input_dim, output_dim, edge_feat, **kwargs) for name in space})
        
        self.bns = torch.nn.ModuleDict({
            name: (
                torch.nn.BatchNorm1d(input_dim if self.pos == 1 else output_dim, track_running_stats=track) 
                if self.bn else Identity()
            ) 
        for name in space})

    def forward(self, x, edge_index, key):
        x_origin = x
        if self.pos == 1:
            # norm - act - drop - conv - (possible) res
            if isinstance(x, tuple):
                x = (self.bns[key](x[0]), self.bns[key](x[1]))
                x = (self.act(x[0]), self.act(x[1]))
                x = (self.dropout(x[0]), self.dropout(x[1]))
            else:
                x = self.bns[key](x)
                x = self.act(x)
                x = self.dropout(x)
        
        x = self.core[key](x, edge_index)

        if self.pos == 2:
            # conv - norm - act - drop - res
            x = self.bns[key](x)
            x = self.act(x)
            x = self.dropout(x)
        
        if self.res:
            if isinstance(x_origin, tuple): x = x + x_origin[1]
            else: x = x + x_origin
        
        return x

    def reset_parameters(self):
        for key in self.core:
            self.core[key].reset_parameters()
            if hasattr(self.bns[key], "reset_parameters"):
                self.bns[key].reset_parameters()


class Conv(torch.nn.Module):
    def __init__(self, space, input_dim, output_dim, pos=0, act=Identity(), dropout=0.0, bn=False, res=False, track=False, **kwargs) -> None:
        super().__init__()
        self.pos = pos
        self.act = act
        self.bn = bn
        if dropout > 0:
            self.dropout = torch.nn.Dropout(p=dropout)
        else:
            self.dropout = Identity()
        self.space = space
        self.res = res

        # span a space
        self.core = torch.nn.ModuleDict({name: op(name, input_dim, output_dim, **kwargs) for name in space})
        
        self.bns = torch.nn.ModuleDict({
            name: (
                torch.nn.BatchNorm1d(input_dim if self.pos == 1 else output_dim, track_running_stats=track) 
                if self.bn else Identity()
            ) 
        for name in space})

    def forward(self, x, edge_index, key):
        x_origin = x
        if self.pos == 1:
            # norm - act - drop - conv - (possible) res
            if isinstance(x, tuple):
                x = (self.bns[key](x[0]), self.bns[key](x[1]))
                x = (self.act(x[0]), self.act(x[1]))
                x = (self.dropout(x[0]), self.dropout(x[1]))
            else:
                x = self.bns[key](x)
                x = self.act(x)
                x = self.dropout(x)
        
        x = self.core[key](x, edge_index)

        if self.pos == 2:
            # conv - norm - act - drop - res
            x = self.bns[key](x)
            x = self.act(x)
            x = self.dropout(x)
        
        if self.res:
            if isinstance(x_origin, tuple): x = x + x_origin[1]
            else: x = x + x_origin
        
        return x

    def reset_parameters(self):
        for key in self.core:
            self.core[key].reset_parameters()
            if hasattr(self.bns[key], "reset_parameters"):
                self.bns[key].reset_parameters()



from collections import namedtuple
from torch.nn import Module
import torch.nn as nn
import torch

Genotype = namedtuple("Genotype", "normal normal_concat")
Genotype_normal = namedtuple("Genotype_normal", "normal normal_concat")


act_list = ["sigmoid", "tanh", "relu", "linear", "elu"]


class LinearConv(Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(LinearConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.linear = torch.nn.Linear(in_channels, out_channels, bias)

    def forward(self, x, edge_index, edge_weight=None):
        return self.linear(x)

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.in_channels, self.out_channels
        )


class SkipConv(Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(SkipConv, self).__init__()
        self.out_dim = out_channels

    def forward(self, x, edge_index, edge_weight=None):
        return x

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.in_channels, self.out_channels
        )


class ZeroConv(Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(ZeroConv, self).__init__()
        self.out_dim = out_channels

    def forward(self, x, edge_index, edge_weight=None):
        return torch.zeros([x.size(0), self.out_dim]).to(x.device)

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.in_channels, self.out_channels
        )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from collections import OrderedDict

import numpy as np
import torch

_counter = 0

_logger = logging.getLogger(__name__)


def global_mutable_counting():
    """
    A program level counter starting from 1.
    """
    global _counter
    _counter += 1
    return _counter


def _reset_global_mutable_counting():
    """
    Reset the global mutable counting to count from 1. Useful when defining multiple models with default keys.
    """
    global _counter
    _counter = 0


def to_device(obj, device):
    """
    Move a tensor, tuple, list, or dict onto device.
    """
    if torch.is_tensor(obj):
        return obj.to(device)
    if isinstance(obj, tuple):
        return tuple(to_device(t, device) for t in obj)
    if isinstance(obj, list):
        return [to_device(t, device) for t in obj]
    if isinstance(obj, dict):
        return {k: to_device(v, device) for k, v in obj.items()}
    if isinstance(obj, (int, float, str)):
        return obj
    raise ValueError("'%s' has unsupported type '%s'" % (obj, type(obj)))


def to_list(arr):
    if torch.is_tensor(arr):
        return arr.cpu().numpy().tolist()
    if isinstance(arr, np.ndarray):
        return arr.tolist()
    if isinstance(arr, (list, tuple)):
        return list(arr)
    return arr


class AverageMeterGroup:
    """
    Average meter group for multiple average meters.
    """

    def __init__(self):
        self.meters = OrderedDict()

    def update(self, data):
        """
        Update the meter group with a dict of metrics.
        Non-exist average meters will be automatically created.
        """
        for k, v in data.items():
            if k not in self.meters:
                self.meters[k] = AverageMeter(k, ":4f")
            self.meters[k].update(v)

    def __getattr__(self, item):
        return self.meters[item]

    def __getitem__(self, item):
        return self.meters[item]

    def __str__(self):
        return "  ".join(str(v) for v in self.meters.values())

    def summary(self):
        """
        Return a summary string of group data.
        """
        return "  ".join(v.summary() for v in self.meters.values())


class AverageMeter:
    """
    Computes and stores the average and current value.

    Parameters
    ----------
    name : str
        Name to display.
    fmt : str
        Format string to print the values.
    """

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        """
        Reset the meter.
        """
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """
        Update with value and weight.

        Parameters
        ----------
        val : float or int
            The new value to be accounted in.
        n : int
            The weight of the new value.
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)

    def summary(self):
        fmtstr = "{name}: {avg" + self.fmt + "}"
        return fmtstr.format(**self.__dict__)


class StructuredMutableTreeNode:
    """
    A structured representation of a search space.
    A search space comes with a root (with `None` stored in its `mutable`), and a bunch of children in its `children`.
    This tree can be seen as a "flattened" version of the module tree. Since nested mutable entity is not supported yet,
    the following must be true: each subtree corresponds to a ``MutableScope`` and each leaf corresponds to a
    ``Mutable`` (other than ``MutableScope``).

    Parameters
    ----------
    mutable : nni.nas.pytorch.mutables.Mutable
        The mutable that current node is linked with.
    """

    def __init__(self, mutable):
        self.mutable = mutable
        self.children = []

    def add_child(self, mutable):
        """
        Add a tree node to the children list of current node.
        """
        self.children.append(StructuredMutableTreeNode(mutable))
        return self.children[-1]

    def type(self):
        """
        Return the ``type`` of mutable content.
        """
        return type(self.mutable)

    def __iter__(self):
        return self.traverse()

    def traverse(self, order="pre", deduplicate=True, memo=None):
        """
        Return a generator that generates a list of mutables in this tree.

        Parameters
        ----------
        order : str
            pre or post. If pre, current mutable is yield before children. Otherwise after.
        deduplicate : bool
            If true, mutables with the same key will not appear after the first appearance.
        memo : dict
            An auxiliary dict that memorize keys seen before, so that deduplication is possible.

        Returns
        -------
        generator of Mutable
        """
        if memo is None:
            memo = set()
        assert order in ["pre", "post"]
        if order == "pre":
            if self.mutable is not None:
                if not deduplicate or self.mutable.key not in memo:
                    memo.add(self.mutable.key)
                    yield self.mutable
        for child in self.children:
            for m in child.traverse(order=order, deduplicate=deduplicate, memo=memo):
                yield m
        if order == "post":
            if self.mutable is not None:
                if not deduplicate or self.mutable.key not in memo:
                    memo.add(self.mutable.key)
                    yield self.mutable



from collections import namedtuple

Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


NA_PRIMITIVES_o = [
  'sage',
  'gcn',
  'gin',
  'gat',
  'graphconv_mean',  # aggr:add mean max
  'mlp',
]

NA_PRIMITIVES = [
  'gcnmol',
  'ginmol',
  #'gatmol',
  'sagemol',
  #'graphmol',  # aggr:add mean max
  'mlpmol',
]

POOL_PRIMITIVES=[
  'none',
]

POOL_PRIMITIVES_o=[
  'hoppool_1',
  'hoppool_2',
  'hoppool_3',

  'topkpool',
  'mlppool',

  'gappool',

  'sagpool',
  'asappool',
  'sag_graphconv', #GCPOOL

  'none',
]
READOUT_PRIMITIVES = [
  'global_mean',
  'global_max',
  'global_sum',
  'none',
  # 'mean_max',
  'global_att',
'global_sort',#DGCNN
  'set2set',  # a seq2seq method
]
ACT_PRIMITIVES = [
  "sigmoid", "tanh", "relu",
  "softplus", "leaky_relu", "relu6", "elu"
]
LA_PRIMITIVES=[
  'l_max',
  'l_concat',
  'l_lstm',
  'l_sum',
  'l_mean',
]





import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Linear, Sequential, ReLU, Conv1d, ELU, PReLU

from torch_geometric.nn import SAGEConv, GATConv, JumpingKnowledge
from torch_geometric.nn import GCNConv, GINConv,GraphConv,LEConv,SGConv,DenseSAGEConv,DenseGCNConv,DenseGINConv,DenseGraphConv
from torch_geometric.nn import global_add_pool,global_mean_pool,global_max_pool,global_sort_pool,GlobalAttention,Set2Set
from torch_geometric.nn import SAGPooling,TopKPooling,EdgePooling,ASAPooling,dense_diff_pool
from .geniepath import GeniePathLayer
from .pooling_zoo import SAGPool_mix, ASAPooling_mix, TOPKpooling_mix, Hoppooling_mix, Gappool_Mixed
from .agg_zoo import GAT_mix,SAGE_mix,Geolayer_mix, GIN_mix
from torch_geometric.nn.inits import reset
from .op_ogb import *
NA_OPS = {
    #SANE
  'sage': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'sage'),
  'sage_sum': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'sum'),
  'sage_max': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'max'),
  'gcn': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'gcn'),
  'gat': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'gat'),
  'gin': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'gin'),
  'arma': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'arma'),
  'gat_sym': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'gat_sym'),
  'gat_linear': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'linear'),
  'gat_cos': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'cos'),
  'gat_generalized_linear': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'generalized_linear'),
  #'geniepath': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'geniepath'),
  'mlp': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'mlp'),

  'gcnmol': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'gcnmol'),
  'gatmol': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'gatmol'),
  'ginmol': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'ginmol'),
  'sagemol': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'sagemol'),
  'graphmol': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'graphmol'),
  'mlpmol': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'mlpmol'),

  #graph classification:
  'graphconv_add': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'graphconv_add'),
  'graphconv_mean': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'graphconv_mean'),
  'graphconv_max': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'graphconv_max'),
  #'sgc': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'sgc'),
  'leconv': lambda in_dim, out_dim: NaAggregator(in_dim, out_dim, 'leconv'),

}
POOL_OPS = {

  'hoppool_1': lambda hidden,ratio,num_nodes:Pooling_func(hidden,ratio,'hoppool_1',num_nodes=num_nodes),
  'hoppool_2': lambda hidden,ratio,num_nodes:Pooling_func(hidden,ratio,'hoppool_2',num_nodes=num_nodes),
  'hoppool_3': lambda hidden,ratio,num_nodes:Pooling_func(hidden,ratio,'hoppool_3',num_nodes=num_nodes),

  'mlppool': lambda hidden, ratio, num_nodes: Pooling_func(hidden, ratio, 'mlppool', num_nodes=num_nodes),
  'topkpool': lambda hidden, ratio, num_nodes: Pooling_func(hidden, ratio, 'topkpool', num_nodes=num_nodes),

  'gappool': lambda hidden, ratio, num_nodes: Pooling_func(hidden, ratio, 'gappool', num_nodes=num_nodes),

  'asappool': lambda hidden, ratio, num_nodes: Pooling_func(hidden, ratio, 'asappool', num_nodes=num_nodes),
  'sagpool': lambda hidden, ratio, num_nodes: Pooling_func(hidden, ratio, 'sagpool', num_nodes=num_nodes),
  'sag_graphconv': lambda hidden, ratio, num_nodes: Pooling_func(hidden, ratio, 'graphconv', num_nodes=num_nodes),

  'none': lambda hidden,ratio,num_nodes:Pooling_func(hidden,ratio, 'none', num_nodes=num_nodes),
}
READOUT_OPS = {
    "global_mean": lambda hidden :Readout_func('mean', hidden),
    "global_sum": lambda hidden  :Readout_func('add', hidden),
    "global_max": lambda hidden  :Readout_func('max', hidden),
    "none":lambda hidden  :Readout_func('none', hidden),
    'global_att': lambda hidden  :Readout_func('att', hidden),
    'global_sort': lambda hidden  :Readout_func('sort',hidden),
    'set2set': lambda hidden  :Readout_func('set2set',hidden),
}


LA_OPS={
  'l_max': lambda hidden_size, num_layers: LaAggregator('max', hidden_size, num_layers),
  'l_concat': lambda hidden_size, num_layers: LaAggregator('cat', hidden_size, num_layers),
  'l_mean': lambda hidden_size, num_layers: LaAggregator('mean', hidden_size, num_layers),
  'l_sum': lambda hidden_size, num_layers: LaAggregator('sum', hidden_size, num_layers),
  'l_lstm': lambda hidden_size, num_layers: LaAggregator('lstm', hidden_size, num_layers)
  #min/max
}

class NaAggregator(nn.Module):

  def __init__(self, in_dim, out_dim, aggregator):
    super(NaAggregator, self).__init__()
    #aggregator, K = agg_str.split('_')
    self.aggregator = aggregator
    if 'sage' == aggregator:
      # self._op = SAGEConv(in_dim, out_dim, normalize=True)
      self._op = SAGE_mix(in_dim, out_dim)
    if 'gcn' == aggregator:
      self._op = GCNConv(in_dim, out_dim)
    if 'arma' == aggregator:
      self._op = ARMAConv(in_dim, out_dim)
    if 'gat' == aggregator:
      heads = 2
      out_dim /= heads
      self._op = GAT_mix(in_dim, int(out_dim), heads=heads, dropout=0.5)
    if 'gin' == aggregator:
      nn1 = Sequential(Linear(in_dim, out_dim), ELU(), Linear(out_dim, out_dim))
      self._op = GIN_mix(nn1)
    if aggregator in ['gat_sym', 'cos', 'linear', 'generalized_linear']:
      heads = 2
      out_dim /= heads
      self._op = Geolayer_mix(in_dim, int(out_dim), heads=heads, att_type=aggregator, dropout=0.5)
    if aggregator in ['sum', 'max']:
      self._op = Geolayer_mix(in_dim, out_dim, att_type='const', agg_type=aggregator, dropout=0.5)
    if aggregator in ['geniepath']:
      self._op = GeniePathLayer(in_dim, out_dim)
    if aggregator =='sgc':
        self._op = SGConv(in_dim, out_dim)
    if 'graphconv'in aggregator:
      aggr = aggregator.split('_')[-1]
      self._op = GraphConv(in_dim, out_dim, aggr=aggr)
    if aggregator == 'leconv':
        self._op = LEConv(in_dim, out_dim)
    if aggregator =='mlp':
      self._op = Sequential(Linear(in_dim, out_dim), ELU(), Linear(out_dim, out_dim))
    elif aggregator =='gcnmol':
        self._op = GCNConvMol(in_dim)
    elif aggregator =='gatmol':
        self._op = GATConvMol(in_dim)
    elif aggregator =='ginmol':
        self._op = GINConvMol(in_dim)
    elif aggregator =='sagemol':
        self._op = SAGEConvMol(in_dim)
    elif aggregator =='graphmol':
        self._op = GraphConvMol(in_dim)
    elif aggregator =='mlpmol':
        self._op = MLPConvMol(in_dim)

    self.mol = 'mol' in aggregator

  def reset_params(self):
    if self.aggregator == 'mlp':
      reset(self._op)
    else:
      self._op.reset_parameters()

  def forward(self, x, edge_index, edge_weight=None, edge_attr=None):
    if self.aggregator == 'mlp':
      return self._op(x)
    elif self.mol:
      return self._op(x, edge_index, edge_attr=edge_attr)
    else:
      return self._op(x, edge_index, edge_weight=edge_weight)


class LaAggregator(nn.Module):

  def __init__(self, mode, hidden_size, num_layers=3):
    super(LaAggregator, self).__init__()
    self.mode = mode
    if self.mode in ['lstm', 'max', 'cat']:
      self.jump = JumpingKnowledge(mode, channels=hidden_size, num_layers=num_layers)
    if mode == 'cat':
        self.lin = Linear(hidden_size * num_layers, hidden_size)
    else:
        self.lin = Linear(hidden_size, hidden_size)
  def reset_params(self):
    self.lin.reset_parameters()
  def forward(self, xs):
    if self.mode in ['lstm', 'max', 'cat']:
      return self.lin(F.elu(self.jump(xs)))
    elif self.mode =='sum':
      return self.lin(F.elu(torch.stack(xs, dim=-1).sum(dim=-1)))
    elif self.mode =='mean':
      return self.lin(F.elu(torch.stack(xs, dim=-1).mean(dim=-1)))

class Readout_trivial(nn.Module):
  def forward(self, x, batch):
      a = global_mean_pool(x, batch)
      b = global_max_pool(x, batch)
      return torch.cat((a,b), 1)

class Readout_func(nn.Module):
  def __init__(self, readout_op, hidden):

    super(Readout_func, self).__init__()
    self.readout_op = readout_op

    if readout_op == 'mean':
      self.readout = global_mean_pool

    elif readout_op == 'max':
      self.readout = global_max_pool

    elif readout_op == 'add':
      self.readout = global_add_pool

    elif readout_op == 'att':
      self.readout = GlobalAttention(Linear(hidden, 1))

    elif readout_op == 'set2set':
      processing_steps = 2
      self.readout = Set2Set(hidden, processing_steps=processing_steps)
      self.s2s_lin = Linear(hidden*processing_steps, hidden)


    elif readout_op == 'sort':
      self.readout = global_sort_pool
      self.k = 10
      self.sort_conv = Conv1d(hidden, hidden, 5)#kernel size 3, output size: hidden,
      self.sort_lin = Linear(hidden*(self.k-5 + 1), hidden)
    elif readout_op =='mema':
      self.readout = global_mean_pool
      self.lin = Linear(hidden*2, hidden)
    elif readout_op == 'none':
      self.readout = global_mean_pool
    # elif self.readout_op == 'mlp':

  def reset_params(self):
    if self.readout_op =='sort':
      self.sort_conv.reset_parameters()
      self.sort_lin.reset_parameters()
    if self.readout_op in ['set2set', 'att']:
      self.readout.reset_parameters()
    if self.readout_op =='set2set':
      self.s2s_lin.reset_parameters()
    if self.readout_op == 'mema':
      self.lin.reset_parameters()
  def forward(self, x, batch, mask):
    #sparse data
    if self.readout_op == 'none':
      x = self.readout(x, batch)
      return x.mul(0.)
      # return None
    elif self.readout_op == 'sort':
      x = self.readout(x, batch, self.k)
      x = x.view(len(x), self.k, -1).permute(0, 2, 1)
      x = F.elu(self.sort_conv(x))
      x = x.view(len(x), -1)
      x = self.sort_lin(x)
      return x
    elif self.readout_op == 'mema':
      x1 = global_mean_pool(x, batch)
      x2 = global_max_pool(x, batch)
      x = torch.cat([x1, x2], dim=-1)
      x = self.lin(x)
      return x
    else:
      x = self.readout(x, batch)
      if self.readout_op == 'set2set':
        x = self.s2s_lin(x)
      return x


class Pooling_trivial(nn.Module):
    def __init__(self, in_channels, ratio=0.5, gnn_type='gcn'):
        super().__init__()
        self.op = SAGPooling(in_channels, ratio=ratio, GNN=GCNConv)

    def forward(self, x, edge_index, edge_weight=None, edge_attr=None, batch=None, attn=None, add_self_loop=False, remove_self_loop=False, ft=False):
        x, edge_index, edge_attr, batch, perm, score = self.op(x, edge_index, None, batch, attn)
        #print(x.size(), batch.size())
        return x, edge_index, edge_attr, batch, perm

class Pooling_func(nn.Module):
  def __init__(self, hidden, ratio, op, dropout=0.6, num_nodes=0):
    super(Pooling_func, self).__init__()
    self.op = op
    self.max_num_nodes = num_nodes
    if op =='sagpool':
      self._op = SAGPool_mix(hidden, ratio=ratio, gnn_type='gcn')
    elif op =='mlppool':
      self._op = SAGPool_mix(hidden, ratio=ratio, gnn_type='mlp')
    elif op =='graphconv':
      self._op = SAGPool_mix(hidden, ratio=ratio, gnn_type='graphconv')

    elif 'hop' in op:
      hop_num = int(op.split('_')[-1])
      self._op = Hoppooling_mix(hidden, ratio=ratio, walk_length=hop_num)
    elif op == 'gappool':
      self._op = Gappool_Mixed(hidden, ratio=ratio)
    elif op == 'topkpool':
      # self._op = TopKPooling(hidden, ratio)
      self._op = TOPKpooling_mix(hidden, ratio=ratio)

    elif op == 'asappool':
      # self._op = ASAPooling(hidden, ratio, dropout=dropout)
      self._op = ASAPooling_mix(hidden, ratio=ratio, dropout=dropout)
  def reset_params(self):
    if self.op != 'none':
      self._op.reset_parameters()

  def forward(self, x, edge_index, edge_weights, data, batch, mask, ft=False):
    if self.op == 'none':
      perm = torch.ones(x.size(0), dtype=torch.float64, device=x.get_device())
      return x, edge_index, edge_weights, batch, perm

    elif self.op in ['asappool', 'topkpool', 'sagpool', 'mlppool', 'hoppool_1', 'hoppool_2', 'hoppool_3', 'gappool', 'graphconv']:
      # print('operations:', self.op)
      x, edge_index, edge_weight, batch, perm = self._op(x=x, edge_index=edge_index, edge_weight=edge_weights, batch=batch, ft=ft)
      return x, edge_index, edge_weight, batch, perm




import torch
from torch import nn
import torch.nn.functional as F
import dgl
from dgl.nn.pytorch.conv import *
import dgl.function as fn
import math


class LinearConv(nn.Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(LinearConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.linear = torch.nn.Linear(in_channels, out_channels, bias)

    def forward(self, g, x, *args, **kwargs):
        return self.linear(x)

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.in_channels, self.out_channels
        )


class ZeroConv(nn.Module):
    def forward(self, g, x, *args, **kwargs):
        out = torch.zeros_like(x)
        out.requires_grad = True
        return out

    def __repr__(self):
        return "ZeroConv()"


class Identity(nn.Module):
    def forward(self, g, x, *args, **kwargs):
        return x

    def __repr__(self):
        return "Identity()"


def glorot(tensor):
    if tensor is not None:
        stdv = math.sqrt(6.0 / (tensor.size(-2) + tensor.size(-1)))
        tensor.data.uniform_(-stdv, stdv)


def zeros(tensor):
    if tensor is not None:
        tensor.data.fill_(0)


class ARMAConv(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        num_stacks=1,
        num_layers=1,
        activation=None,
        dropout=0.0,
        bias=True,
    ):
        super(ARMAConv, self).__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.K = num_stacks
        self.T = num_layers
        self.activation = activation
        self.dropout = nn.Dropout(p=dropout)

        # init weight
        self.w_0 = nn.ModuleDict(
            {str(k): nn.Linear(in_dim, out_dim, bias=False) for k in range(self.K)}
        )
        # deeper weight
        self.w = nn.ModuleDict(
            {str(k): nn.Linear(out_dim, out_dim, bias=False) for k in range(self.K)}
        )
        # v
        self.v = nn.ModuleDict(
            {str(k): nn.Linear(in_dim, out_dim, bias=False) for k in range(self.K)}
        )
        # bias
        if bias:
            self.bias = nn.Parameter(torch.Tensor(self.K, self.T, 1, self.out_dim))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self):
        for k in range(self.K):
            glorot(self.w_0[str(k)].weight)
            glorot(self.w[str(k)].weight)
            glorot(self.v[str(k)].weight)
        zeros(self.bias)

    def forward(self, g, feats):
        with g.local_scope():
            init_feats = feats
            # assume that the graphs are undirected and graph.in_degrees() is the same as graph.out_degrees()
            degs = g.in_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5).to(feats.device).unsqueeze(1)
            output = None

            for k in range(self.K):
                feats = init_feats
                for t in range(self.T):
                    feats = feats * norm
                    g.ndata["h"] = feats
                    g.update_all(fn.copy_u("h", "m"), fn.sum("m", "h"))
                    feats = g.ndata.pop("h")
                    feats = feats * norm

                    if t == 0:
                        feats = self.w_0[str(k)](feats)
                    else:
                        feats = self.w[str(k)](feats)

                    feats += self.dropout(self.v[str(k)](init_feats))
                    feats += self.v[str(k)](self.dropout(init_feats))

                    if self.bias is not None:
                        feats += self.bias[k][t]

                    if self.activation is not None:
                        feats = self.activation(feats)

                if output is None:
                    output = feats
                else:
                    output += feats

            return output / self.K


class GATConvC(GATConv):
    def __init__(
        self,
        in_feats,
        out_feats,
        num_heads,
        feat_drop=0.0,
        attn_drop=0.0,
        negative_slope=0.2,
        residual=False,
        activation=None,
        allow_zero_in_degree=False,
        bias=True,
        concat=False,
    ):
        super(GATConvC, self).__init__(
            in_feats,
            out_feats,
            num_heads,
            feat_drop,
            attn_drop,
            negative_slope,
            residual,
            activation,
            allow_zero_in_degree,
            bias,
        )
        self.concat = concat

    def forward(self, graph, feat, get_attention=False):
        x = super().forward(graph, feat, get_attention=get_attention)
        if get_attention:
            x = x[0]
        out = x
        if self.concat:
            out = out.view(-1, self._num_heads * self._out_feats)
        else:
            out = out.mean(dim=1)
        return out


class MChebConv(ChebConv):
    def __init__(self, in_feats, out_feats, k, activation=F.relu, bias=True):
        super().__init__(in_feats, out_feats, k, activation=activation, bias=bias)

    def forward(self, graph, feat):
        lambda_max = dgl.laplacian_lambda_max(graph)  # may be very slow
        return super().forward(graph, feat, lambda_max=lambda_max)


def gnn_map(gnn_name, in_dim, out_dim, concat=False, bias=True) -> nn.Module:
    """

    :param gnn_name:
    :param in_dim:
    :param out_dim:
    :param concat: for gat, concat multi-head output or not
    :return: GNN model
    """
    # now gat,sage,chebconv may be different from pyg
    if gnn_name == "gat_8":
        return GATConvC(in_dim, out_dim, 8, concat=concat, bias=bias)
    elif gnn_name == "gat_6":
        return GATConvC(in_dim, out_dim, 6, concat=concat, bias=bias)
    elif gnn_name == "gat_4":
        return GATConvC(in_dim, out_dim, 4, concat=concat, bias=bias)
    elif gnn_name == "gat_2":
        return GATConvC(in_dim, out_dim, 2, concat=concat, bias=bias)
    elif gnn_name in ["gat_1", "gat"]:
        return GATConvC(in_dim, out_dim, 1, concat=concat, bias=bias)
    elif gnn_name == "gcn":
        return GraphConv(in_dim, out_dim)
    elif gnn_name == "cheb":
        return MChebConv(in_dim, out_dim, k=2, bias=bias)
    elif gnn_name == "sage":
        return SAGEConv(in_dim, out_dim, bias=bias, aggregator_type="pool")
    elif gnn_name == "gated":
        return GatedGraphConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "arma":
        return ARMAConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "sg":
        return SGConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "linear":
        return LinearConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "zero":
        return ZeroConv()
    elif gnn_name == "identity":
        return Identity()
    elif hasattr(dgl.nn.conv, gnn_name):
        cls = getattr(dgl.nn.conv, gnn_name)
        assert isinstance(cls, type), "Only support modules, get %s" % (gnn_name)
        kwargs = {
            "in_channels": in_dim,
            "out_channels": out_dim,
            "concat": concat,
            "bias": bias,
        }
        kwargs = {
            key: kwargs[key]
            for key in cls.__init__.__code__.co_varnames
            if key in kwargs
        }
        return cls(**kwargs)
    raise KeyError("Cannot parse key %s" % (gnn_name))


from dgl.utils import expand_as_pair
from dgl.nn.functional import edge_softmax
from torch.nn import Parameter

import torch
from torch import Tensor
# from ..scatter_utils import *
from typing import Optional, Tuple, Union

OptTensor = Optional[Tensor]


def softmax(
    src: Tensor,
    index: Optional[Tensor] = None,
    ptr: Optional[Tensor] = None,
    num_nodes: Optional[int] = None,
    dim: int = 0,
) -> Tensor:
    if ptr is not None:
        dim = dim + src.dim() if dim < 0 else dim
        size = ([1] * dim) + [-1]
        ptr = ptr.view(size)
        src_max = gather_csr(segment_csr(src, ptr, reduce="max"), ptr)
        out = (src - src_max).exp()
        out_sum = gather_csr(segment_csr(out, ptr, reduce="sum"), ptr)
    elif index is not None:
        N = maybe_num_nodes(index, num_nodes)
        src_max = scatter(src, index, dim, dim_size=N, reduce="max")
        src_max = src_max.index_select(dim, index)
        out = (src - src_max).exp()
        out_sum = scatter(out, index, dim, dim_size=N, reduce="sum")
        out_sum = out_sum.index_select(dim, index)
    else:
        raise NotImplementedError

    return out / (out_sum + 1e-16)


def maybe_num_nodes(edge_index, num_nodes=None):
    if num_nodes is not None:
        return num_nodes
    elif isinstance(edge_index, Tensor):
        return int(edge_index.max()) + 1 if edge_index.numel() > 0 else 0
    else:
        return max(edge_index.size(0), edge_index.size(1))


def add_remaining_self_loops(
    edge_index: Tensor,
    edge_attr: OptTensor = None,
    fill_value: Union[float, Tensor, str] = None,
    num_nodes: Optional[int] = None,
) -> Tuple[Tensor, OptTensor]:
    r"""Adds remaining self-loop :math:`(i,i) \in \mathcal{E}` to every node
    :math:`i \in \mathcal{V}` in the graph given by :attr:`edge_index`.
    In case the graph is weighted or has multi-dimensional edge features
    (:obj:`edge_attr != None`), edge features of non-existing self-loops will
    be added according to :obj:`fill_value`.

    Args:
        edge_index (LongTensor): The edge indices.
        edge_attr (Tensor, optional): Edge weights or multi-dimensional edge
            features. (default: :obj:`None`)
        fill_value (float or Tensor or str, optional): The way to generate
            edge features of self-loops (in case :obj:`edge_attr != None`).
            If given as :obj:`float` or :class:`torch.Tensor`, edge features of
            self-loops will be directly given by :obj:`fill_value`.
            If given as :obj:`str`, edge features of self-loops are computed by
            aggregating all features of edges that point to the specific node,
            according to a reduce operation. (:obj:`"add"`, :obj:`"mean"`,
            :obj:`"min"`, :obj:`"max"`, :obj:`"mul"`). (default: :obj:`1.`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)

    :rtype: (:class:`LongTensor`, :class:`Tensor`)
    """
    N = maybe_num_nodes(edge_index, num_nodes)
    mask = edge_index[0] != edge_index[1]

    loop_index = torch.arange(0, N, dtype=torch.long, device=edge_index.device)
    loop_index = loop_index.unsqueeze(0).repeat(2, 1)

    if edge_attr is not None:
        if fill_value is None:
            loop_attr = edge_attr.new_full((N,) + edge_attr.size()[1:], 1.0)

        elif isinstance(fill_value, (int, float)):
            loop_attr = edge_attr.new_full((N,) + edge_attr.size()[1:], fill_value)
        elif isinstance(fill_value, Tensor):
            loop_attr = fill_value.to(edge_attr.device, edge_attr.dtype)
            if edge_attr.dim() != loop_attr.dim():
                loop_attr = loop_attr.unsqueeze(0)
            sizes = [N] + [1] * (loop_attr.dim() - 1)
            loop_attr = loop_attr.repeat(*sizes)

        elif isinstance(fill_value, str):
            loop_attr = scatter(
                edge_attr, edge_index[1], dim=0, dim_size=N, reduce=fill_value
            )
        else:
            raise AttributeError("No valid 'fill_value' provided")

        inv_mask = ~mask
        loop_attr[edge_index[0][inv_mask]] = edge_attr[inv_mask]

        edge_attr = torch.cat([edge_attr[mask], loop_attr], dim=0)

    edge_index = torch.cat([edge_index[:, mask], loop_index], dim=1)
    return edge_index, edge_attr


class GeoLayer(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        heads=1,
        concat=True,
        negative_slope=0.2,
        dropout=0,
        bias=True,
        att_type="gat",
        agg_type="sum",
        pool_dim=0,
    ):
        super(GeoLayer, self).__init__()
        if agg_type in ["sum", "mlp"]:
            agg_type = "add"
        elif agg_type in ["mean", "max"]:
            pass
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.att_type = att_type
        self.agg_type = agg_type
        self.reduce_func = {"add": fn.sum, "mean": fn.mean, "max": fn.max}.get(
            self.agg_type, None
        )

        # GCN weight
        self.gcn_weight = None

        self.weight = Parameter(torch.Tensor(in_channels, heads * out_channels))
        self.att = Parameter(torch.Tensor(1, heads, 2 * out_channels))

        if bias and concat:
            self.bias = Parameter(torch.Tensor(heads * out_channels))
        elif bias and not concat:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        if self.att_type in ["generalized_linear"]:
            self.general_att_layer = torch.nn.Linear(out_channels, 1, bias=False)

        if self.agg_type in ["mean", "max", "mlp"]:
            if pool_dim <= 0:
                pool_dim = 128
        self.pool_dim = pool_dim
        if pool_dim != 0:
            self.pool_layer = torch.nn.ModuleList()
            self.pool_layer.append(torch.nn.Linear(self.out_channels, self.pool_dim))
            self.pool_layer.append(torch.nn.Linear(self.pool_dim, self.out_channels))
        else:
            pass
        self.reset_parameters()

    @staticmethod
    def norm(edge_index, num_nodes, edge_weight, improved=False, dtype=None):
        if edge_weight is None:
            edge_weight = torch.ones(
                (edge_index.size(1),), dtype=dtype, device=edge_index.device
            )

        fill_value = 1 if not improved else 2
        edge_index, edge_weight = add_remaining_self_loops(
            edge_index, edge_weight, fill_value, num_nodes
        )

        row, col = edge_index
        deg = scatter_add(edge_weight, row, dim=0, dim_size=num_nodes)
        deg_inv_sqrt = deg.pow(-0.5)
        deg_inv_sqrt[deg_inv_sqrt == float("inf")] = 0

        return edge_index, deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.att)
        zeros(self.bias)

        if self.att_type in ["generalized_linear"]:
            glorot(self.general_att_layer.weight)

        if self.pool_dim != 0:
            for layer in self.pool_layer:
                glorot(layer.weight)
                zeros(layer.bias)

    def forward(self, graph, feat):
        """"""
        # add self_loop
        graph = dgl.remove_self_loop(graph)
        graph = dgl.add_self_loop(graph)
        # prepare
        x = feat
        x = torch.mm(x, self.weight).view(-1, self.heads, self.out_channels)
        out = self.propagate(graph, x)
        return out

    def propagate(self, graph, feat):
        # x_i torch.Size([13264, 2, 4])
        # x_j torch.Size([13264, 2, 4])
        # edge_index torch.Size([2, 13264])
        # num_nodes 2708
        # x_i is target; x_j is source
        edge_index = torch.stack(graph.edges())
        x_i = torch.index_select(feat, 0, edge_index[1])
        x_j = torch.index_select(feat, 0, edge_index[0])
        num_nodes = graph.num_nodes()
        if self.att_type == "const":  # Const OK
            if self.training and self.dropout > 0:
                x_j = F.dropout(x_j, p=self.dropout, training=True)
            neighbor = x_j
        elif self.att_type == "gcn":  # GCN OK
            if self.gcn_weight is None or self.gcn_weight.size(0) != x_j.size(
                0
            ):  # calculate norm by degree like GCN
                _, norm = self.norm(edge_index, num_nodes, None)
                self.gcn_weight = norm
            neighbor = self.gcn_weight.view(-1, 1, 1) * x_j
        else:  # Attention OK
            # Compute attention coefficients.
            alpha = self.apply_attention(edge_index, num_nodes, x_i, x_j)
            alpha = softmax(alpha, edge_index[0], num_nodes=num_nodes)
            # Sample attention coefficients stochastically.
            if self.training and self.dropout > 0:
                alpha = F.dropout(alpha, p=self.dropout, training=True)

            neighbor = x_j * alpha.view(-1, self.heads, 1)

        if self.pool_dim > 0:
            # neighbor torch.Size([13264, 2, 4])
            for layer in self.pool_layer:
                neighbor = layer(neighbor)

        graph.edata["e"] = neighbor
        # aggregate : need neighbor as edge, ok
        if self.reduce_func is not None:
            graph.update_all(fn.copy_e("e", "e"), self.reduce_func("e", "h"))
            out = graph.ndata["h"]
            graph.edata.pop("e")
        else:
            out = neighbor
            pass

        out = self.update(out)
        return out

    def apply_attention(self, edge_index, num_nodes, x_i, x_j):
        if self.att_type == "gat":
            alpha = (torch.cat([x_i, x_j], dim=-1) * self.att).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope)

        elif self.att_type == "gat_sym":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            alpha = (x_i * wl).sum(dim=-1) + (x_j * wr).sum(dim=-1)
            alpha_2 = (x_j * wl).sum(dim=-1) + (x_i * wr).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope) + F.leaky_relu(
                alpha_2, self.negative_slope
            )

        elif self.att_type == "linear":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            al = x_j * wl
            ar = x_j * wr
            alpha = al.sum(dim=-1) + ar.sum(dim=-1)
            alpha = torch.tanh(alpha)
        elif self.att_type == "cos":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            alpha = x_i * wl * x_j * wr
            alpha = alpha.sum(dim=-1)

        elif self.att_type == "generalized_linear":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            al = x_i * wl
            ar = x_j * wr
            alpha = al + ar
            alpha = torch.tanh(alpha)
            alpha = self.general_att_layer(alpha)
        else:
            raise Exception("Wrong attention type:", self.att_type)
        return alpha

    def update(self, aggr_out):  # torch.Size([2708, 2, 4])
        if self.concat is True:
            aggr_out = aggr_out.view(
                -1, self.heads * self.out_channels
            )  # torch.Size([2708, 8])
        else:
            aggr_out = aggr_out.mean(dim=1)

        if self.bias is not None:
            aggr_out = aggr_out + self.bias
        return aggr_out

    def __repr__(self):
        return "{}({}, {}, heads={})".format(
            self.__class__.__name__, self.in_channels, self.out_channels, self.heads
        )

    def get_param_dict(self):
        params = {}
        key = f"{self.att_type}_{self.agg_type}_{self.in_channels}_{self.out_channels}_{self.heads}"
        weight_key = key + "_weight"
        att_key = key + "_att"
        agg_key = key + "_agg"
        bais_key = key + "_bais"

        params[weight_key] = self.weight
        params[att_key] = self.att
        params[bais_key] = self.bias
        if hasattr(self, "pool_layer"):
            params[agg_key] = self.pool_layer.state_dict()

        return params

    def load_param(self, params):
        key = f"{self.att_type}_{self.agg_type}_{self.in_channels}_{self.out_channels}_{self.heads}"
        weight_key = key + "_weight"
        att_key = key + "_att"
        agg_key = key + "_agg"
        bais_key = key + "_bais"

        if weight_key in params:
            self.weight = params[weight_key]

        if att_key in params:
            self.att = params[att_key]

        if bais_key in params:
            self.bias = params[bais_key]

        if agg_key in params and hasattr(self, "pool_layer"):
            self.pool_layer.load_state_dict(params[agg_key])



import torch
import typing as _typ
import torch.nn as nn
import torch.nn.functional as F

from .base import BaseSpace
from .operation import map_nn, act_map, GeoLayer
from autogllight.utils.backend import BackendOperator as BK

DEFAULT_GNN = [
    "gat",
    "gcn",
    "cos",
    "const",
    "gat_sym",
    "linear",
    "generalized_linear",
]
DEFAULT_AGG = [
    "sum",
    "mean",
    "max",
    "mlp",
]

DEFAULT_ACT = [
    "sigmoid",
    "tanh",
    "relu",
    "linear",
    "softplus",
    "leaky_relu",
    "relu6",
    "elu",
]
DEFAULT_HEAD = [1, 2, 4, 6, 8, 16]
DEFAULT_OUT = [4, 8, 16, 32, 64, 128, 256]


class GraphNasMacroNodeClassificationSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.6,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        gnns: _typ.Tuple = DEFAULT_GNN,
        aggs=DEFAULT_AGG,
        acts=DEFAULT_ACT,
        heads=DEFAULT_HEAD,
        outs=DEFAULT_OUT,
        search_act_con=False,
    ):
        super().__init__()
        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.gnns = gnns
        self.aggs = aggs
        self.acts = acts
        self.heads = heads
        self.outs = outs
        self.dropout = dropout
        self.search_act_con = search_act_con

    def build_graph(self):
        layer_nums = self.layer_number
        state_num = 5

        # build hidden layer
        for i in range(layer_nums):
            # message choice
            layer = i * state_num + 0
            op_candidates = map_nn(self.gnns)
            key = f"attention_{i}"
            self.setLayerChoice(layer, op_candidates, key=key)

            # aggregation choice
            layer = i * state_num + 1
            op_candidates = map_nn(self.aggs)
            key = f"aggregator_{i}"
            self.setLayerChoice(layer, op_candidates, key=key)

            # activation choice
            layer = i * state_num + 0
            op_candidates = map_nn(self.acts)
            key = f"act_{i}"
            self.setLayerChoice(layer, op_candidates, key=key)

            # head choice
            layer = i * state_num + 0
            op_candidates = map_nn(self.heads)
            key = f"head_{i}"
            self.setLayerChoice(layer, op_candidates, key=key)

            # out choice
            if i < layer_nums - 1:
                layer = i * state_num + 0
                op_candidates = map_nn(self.outs)
                key = f"out_channels_{i}"
                self.setLayerChoice(layer, op_candidates, key=key)

    def parse_model(self, selection):
        sel_list = []
        for i in range(self.layer_number):
            sel_list.append(self.gnns[selection[f"attention_{i}"]])
            sel_list.append(self.aggs[selection[f"aggregator_{i}"]])
            sel_list.append(self.acts[selection[f"act_{i}"]])
            sel_list.append(self.heads[selection[f"head_{i}"]])
            if i < self.layer_number - 1:
                sel_list.append(self.outs[selection[f"out_channels_{i}"]])
        sel_list.append(self.output_dim)
        # sel_list = ['const', 'sum', 'relu6', 2, 128, 'gat', 'sum', 'linear', 2, 7]
        model = GraphNet(
            sel_list,
            self.input_dim,
            self.output_dim,
            self.dropout,
            multi_label=False,
            batch_normal=False,
            layers=self.layer_number,
        ).wrap()
        return model


class GraphNet(BaseSpace):
    def __init__(
        self,
        actions,
        num_feat,
        num_label,
        drop_out=0.6,
        multi_label=False,
        batch_normal=True,
        state_num=5,
        residual=False,
        layers=2,
    ):
        self.residual = residual
        self.batch_normal = batch_normal
        self.layer_nums = layers
        self.multi_label = multi_label
        self.num_feat = num_feat
        self.num_label = num_label
        self.input_dim = num_feat
        self.output_dim = num_label
        self.dropout = drop_out

        super().__init__()
        self.build_model(
            actions, batch_normal, drop_out, num_feat, num_label, state_num
        )

    def build_model(
        self, actions, batch_normal, drop_out, num_feat, num_label, state_num
    ):
        if self.residual:
            self.fcs = torch.nn.ModuleList()
        if self.batch_normal:
            self.bns = torch.nn.ModuleList()
        self.layers = torch.nn.ModuleList()
        self.acts = []
        self.gates = torch.nn.ModuleList()
        self.build_hidden_layers(
            actions,
            batch_normal,
            drop_out,
            self.layer_nums,
            num_feat,
            num_label,
            state_num,
        )

    def build_hidden_layers(
        self,
        actions,
        batch_normal,
        drop_out,
        layer_nums,
        num_feat,
        num_label,
        state_num=6,
    ):

        # build hidden layer
        for i in range(layer_nums):

            if i == 0:
                in_channels = num_feat
            else:
                in_channels = out_channels * head_num

            # extract layer information
            attention_type = actions[i * state_num + 0]
            aggregator_type = actions[i * state_num + 1]
            act = actions[i * state_num + 2]
            head_num = actions[i * state_num + 3]
            out_channels = actions[i * state_num + 4]
            concat = True
            if i == layer_nums - 1:
                concat = False
            if self.batch_normal:
                self.bns.append(torch.nn.BatchNorm1d(in_channels, momentum=0.5))
            self.layers.append(
                GeoLayer(
                    in_channels,
                    out_channels,
                    head_num,
                    concat,
                    dropout=self.dropout,
                    att_type=attention_type,
                    agg_type=aggregator_type,
                )
            )
            self.acts.append(act_map(act))
            if self.residual:
                if concat:
                    self.fcs.append(
                        torch.nn.Linear(in_channels, out_channels * head_num)
                    )
                else:
                    self.fcs.append(torch.nn.Linear(in_channels, out_channels))

    def forward(self, data):
        output = BK.feat(data)
        # output, edge_index_all = data.x, data.edge_index  # x [2708,1433] ,[2, 10556]
        if self.residual:
            for i, (act, layer, fc) in enumerate(zip(self.acts, self.layers, self.fcs)):
                output = F.dropout(output, p=self.dropout, training=self.training)
                if self.batch_normal:
                    output = self.bns[i](output)

                # output = act(layer(output, edge_index_all) + fc(output))
                output = act(BK.gconv(layer, data, output) + fc(output))
        else:
            for i, (act, layer) in enumerate(zip(self.acts, self.layers)):
                output = F.dropout(output, p=self.dropout, training=self.training)
                if self.batch_normal:
                    output = self.bns[i](output)
                # output = act(layer(output, edge_index_all))
                output = act(BK.gconv(layer, data, output))

        if not self.multi_label:
            output = F.log_softmax(output, dim=1)
        return output

    def __repr__(self):
        result_lines = ""
        for each in self.layers:
            result_lines += str(each)
        return result_lines

    @staticmethod
    def merge_param(old_param, new_param, update_all):
        for key in new_param:
            if update_all or key not in old_param:
                old_param[key] = new_param[key]
        return old_param

    def get_param_dict(self, old_param=None, update_all=True):
        if old_param is None:
            result = {}
        else:
            result = old_param
        for i in range(self.layer_nums):
            key = "layer_%d" % i
            new_param = self.layers[i].get_param_dict()
            if key in result:
                new_param = self.merge_param(result[key], new_param, update_all)
                result[key] = new_param
            else:
                result[key] = new_param
        if self.residual:
            for i, fc in enumerate(self.fcs):
                key = f"layer_{i}_fc_{fc.weight.size(0)}_{fc.weight.size(1)}"
                result[key] = self.fcs[i]
        if self.batch_normal:
            for i, bn in enumerate(self.bns):
                key = f"layer_{i}_fc_{bn.weight.size(0)}"
                result[key] = self.bns[i]
        return result

    def load_param(self, param):
        if param is None:
            return

        for i in range(self.layer_nums):
            self.layers[i].load_param(param["layer_%d" % i])

        if self.residual:
            for i, fc in enumerate(self.fcs):
                key = f"layer_{i}_fc_{fc.weight.size(0)}_{fc.weight.size(1)}"
                if key in param:
                    self.fcs[i] = param[key]
        if self.batch_normal:
            for i, bn in enumerate(self.bns):
                key = f"layer_{i}_fc_{bn.weight.size(0)}"
                if key in param:
                    self.bns[i] = param[key]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
import logging

from .mutables import InputChoice, LayerChoice, MutableScope
from .mutator import Mutator
from .utils import to_list


_logger = logging.getLogger(__name__)


class FixedArchitecture(Mutator):
    """
    Fixed architecture mutator that always selects a certain graph.

    Parameters
    ----------
    model : nn.Module
        A mutable network.
    fixed_arc : dict
        Preloaded architecture object.
    strict : bool
        Force everything that appears in ``fixed_arc`` to be used at least once.
    verbose : bool
        Print log messages if set to True
    """

    def __init__(self, model, fixed_arc, strict=True, verbose=True):
        super().__init__(model)
        self._fixed_arc = fixed_arc
        self.verbose = verbose

        mutable_keys = set(
            [
                mutable.key
                for mutable in self.mutables
                if not isinstance(mutable, MutableScope)
            ]
        )
        fixed_arc_keys = set(self._fixed_arc.keys())
        if fixed_arc_keys - mutable_keys:
            raise RuntimeError(
                "Unexpected keys found in fixed architecture: {}.".format(
                    fixed_arc_keys - mutable_keys
                )
            )
        if mutable_keys - fixed_arc_keys:
            raise RuntimeError(
                "Missing keys in fixed architecture: {}.".format(
                    mutable_keys - fixed_arc_keys
                )
            )
        self._fixed_arc = self._from_human_readable_architecture(self._fixed_arc)

    def _from_human_readable_architecture(self, human_arc):
        # convert from an exported architecture
        result_arc = {
            k: to_list(v) for k, v in human_arc.items()
        }  # there could be tensors, numpy arrays, etc.
        # First, convert non-list to list, because there could be {"op1": 0} or {"op1": "conv"},
        # which means {"op1": [0, ]} ir {"op1": ["conv", ]}
        result_arc = {
            k: v if isinstance(v, list) else [v] for k, v in result_arc.items()
        }
        # Second, infer which ones are multi-hot arrays and which ones are in human-readable format.
        # This is non-trivial, since if an array in [0, 1], we cannot know for sure it means [false, true] or [true, true].
        # Here, we assume an multihot array has to be a boolean array or a float array and matches the length.
        for mutable in self.mutables:
            if mutable.key not in result_arc:
                continue  # skip silently
            choice_arr = result_arc[mutable.key]
            if all(isinstance(v, bool) for v in choice_arr) or all(
                isinstance(v, float) for v in choice_arr
            ):
                if (
                    isinstance(mutable, LayerChoice) and len(mutable) == len(choice_arr)
                ) or (
                    isinstance(mutable, InputChoice)
                    and mutable.n_candidates == len(choice_arr)
                ):
                    # multihot, do nothing
                    continue
            if isinstance(mutable, LayerChoice):
                choice_arr = [
                    mutable.names.index(val) if isinstance(val, str) else val
                    for val in choice_arr
                ]
                choice_arr = [i in choice_arr for i in range(len(mutable))]
            elif isinstance(mutable, InputChoice):
                choice_arr = [
                    mutable.choose_from.index(val) if isinstance(val, str) else val
                    for val in choice_arr
                ]
                choice_arr = [i in choice_arr for i in range(mutable.n_candidates)]
            result_arc[mutable.key] = choice_arr
        return result_arc

    def sample_search(self):
        """
        Always returns the fixed architecture.
        """
        return self._fixed_arc

    def sample_final(self):
        """
        Always returns the fixed architecture.
        """
        return self._fixed_arc

    def replace_layer_choice(self, module=None, prefix=""):
        """
        Replace layer choices with selected candidates. It's done with best effort.
        In case of weighted choices or multiple choices. if some of the choices on weighted with zero, delete them.
        If single choice, replace the module with a normal module.

        Parameters
        ----------
        module : nn.Module
            Module to be processed.
        prefix : str
            Module name under global namespace.
        """
        if module is None:
            module = self.model
        for name, mutable in module.named_children():
            global_name = (prefix + "." if prefix else "") + name
            if isinstance(mutable, LayerChoice):
                chosen = self._fixed_arc[mutable.key]
                if sum(chosen) == 1 and max(chosen) == 1 and not mutable.return_mask:
                    # sum is one, max is one, there has to be an only one
                    # this is compatible with both integer arrays, boolean arrays and float arrays
                    if self.verbose:
                        _logger.info(
                            "Replacing %s with candidate number %d.",
                            global_name,
                            chosen.index(1),
                        )
                    setattr(module, name, mutable[chosen.index(1)])
                else:
                    if mutable.return_mask and self.verbose:
                        _logger.info(
                            "`return_mask` flag of %s is true. As it relies on the behavior of LayerChoice, "
                            "LayerChoice will not be replaced."
                        )
                    # remove unused parameters
                    for ch, n in zip(chosen, mutable.names):
                        if ch == 0 and not isinstance(ch, float):
                            setattr(mutable, n, None)
            else:
                self.replace_layer_choice(mutable, global_name)


def apply_fixed_architecture(model, fixed_arc, verbose=True):
    """
    Load architecture from `fixed_arc` and apply to model.

    Parameters
    ----------
    model : torch.nn.Module
        Model with mutables.
    fixed_arc : str or dict
        Path to the JSON that stores the architecture, or dict that stores the exported architecture.
    verbose : bool
        Print log messages if set to True

    Returns
    -------
    FixedArchitecture
        Mutator that is responsible for fixes the graph.
    """

    if isinstance(fixed_arc, str):
        with open(fixed_arc) as f:
            fixed_arc = json.load(f)
    architecture = FixedArchitecture(model, fixed_arc, verbose)
    architecture.reset()

    # for the convenience of parameters counting
    architecture.replace_layer_choice()
    return architecture



import argparse
import os.path as osp

import torch
from torch_geometric.datasets import PPI
from torch_geometric.data import DataLoader
from torch_geometric.nn import GATConv
from sklearn.metrics import f1_score

#parser = argparse.ArgumentParser()
#parser.add_argument('--model', type=str, default='GeniePathLazy')
#args = parser.parse_args()
#assert args.model in ['GeniePath', 'GeniePathLazy']

#path = osp.join(osp.dirname(osp.realpath(__file__)), 'data', 'PPI')
#train_dataset = PPI(path, split='train')
#val_dataset = PPI(path, split='val')
#test_dataset = PPI(path, split='test')
#train_loader = DataLoader(train_dataset, batch_size=1, shuffle=True)
#val_loader = DataLoader(val_dataset, batch_size=2, shuffle=False)
#test_loader = DataLoader(test_dataset, batch_size=2, shuffle=False)
#
#dim = 256
#lstm_hidden = 256
#layer_num = 4


class Breadth(torch.nn.Module):
    def __init__(self, in_dim, out_dim):
        super(Breadth, self).__init__()
        self.gatconv = GATConv(in_dim, out_dim, heads=1)
    def reset_parameters(self):
        self.gatconv.reset_parameters()
    def forward(self, x, edge_index):
        x = torch.tanh(self.gatconv(x, edge_index))
        return x


class Depth(torch.nn.Module):
    def __init__(self, in_dim, hidden):
        super(Depth, self).__init__()
        self.lstm = torch.nn.LSTM(in_dim, hidden, 1, bias=False)
    def reset_parameters(self):
        self.lstm.reset_parameters()
    def forward(self, x, h, c):
        x, (h, c) = self.lstm(x, (h, c))
        return x, (h, c)


class GeniePathLayer(torch.nn.Module):
    def __init__(self, in_dim, hidden):
        super(GeniePathLayer, self).__init__()
        self.breadth_func = Breadth(in_dim, hidden)
        self.depth_func = Depth(hidden, hidden)
        self.in_dim = in_dim
        self.hidden = hidden
        self.lstm_hidden = hidden

    def reset_parameters(self):
        self.breadth_func.reset_parameters()
        self.depth_func.reset_parameters()

    def forward(self, x, edge_index):
        h = torch.zeros(1, x.shape[0], self.lstm_hidden, device=x.device)
        c = torch.zeros(1, x.shape[0], self.lstm_hidden, device=x.device)
        x = self.breadth_func(x, edge_index)
        x = x[None, :]
        x, (h, c) = self.depth_func(x, h, c)
        x = x[0]
        return x


class GeniePath(torch.nn.Module):
    def __init__(self, in_dim, out_dim):
        super(GeniePath, self).__init__()
        self.lin1 = torch.nn.Linear(in_dim, dim)
        self.gplayers = torch.nn.ModuleList(
            [GeniePathLayer(dim) for i in range(layer_num)])
        self.lin2 = torch.nn.Linear(dim, out_dim)

    def forward(self, x, edge_index):
        x = self.lin1(x)
        h = torch.zeros(1, x.shape[0], lstm_hidden, device=x.device)
        c = torch.zeros(1, x.shape[0], lstm_hidden, device=x.device)
        for i, l in enumerate(self.gplayers):
            x, (h, c) = self.gplayers[i](x, edge_index, h, c)
        x = self.lin2(x)
        return x


class GeniePathLazy(torch.nn.Module):
    def __init__(self, in_dim, out_dim):
        super(GeniePathLazy, self).__init__()
        self.lin1 = torch.nn.Linear(in_dim, dim)
        self.breadths = torch.nn.ModuleList(
            [Breadth(dim, dim) for i in range(layer_num)])
        self.depths = torch.nn.ModuleList(
            [Depth(dim * 2, lstm_hidden) for i in range(layer_num)])
        self.lin2 = torch.nn.Linear(dim, out_dim)

    def forward(self, x, edge_index):
        x = self.lin1(x)
        h = torch.zeros(1, x.shape[0], lstm_hidden, device=x.device)
        c = torch.zeros(1, x.shape[0], lstm_hidden, device=x.device)
        h_tmps = []
        for i, l in enumerate(self.breadths):
            h_tmps.append(self.breadths[i](x, edge_index))
        x = x[None, :]
        for i, l in enumerate(self.depths):
            in_cat = torch.cat((h_tmps[i][None, :], x), -1)
            x, (h, c) = self.depths[i](in_cat, h, c)
        x = self.lin2(x[0])
        return x


#device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
#kwargs = {'GeniePath': GeniePath, 'GeniePathLazy': GeniePathLazy}
#model = kwargs[args.model](train_dataset.num_features,
#                           train_dataset.num_classes).to(device)
#loss_op = torch.nn.BCEWithLogitsLoss()
#optimizer = torch.optim.Adam(model.parameters(), lr=0.005)


def train():
    model.train()

    total_loss = 0
    for data in train_loader:
        num_graphs = data.num_graphs
        data.batch = None
        data = data.to(device)
        optimizer.zero_grad()
        loss = loss_op(model(data.x, data.edge_index), data.y)
        total_loss += loss.item() * num_graphs
        loss.backward()
        optimizer.step()
    return total_loss / len(train_loader.dataset)


def test(loader):
    model.eval()

    ys, preds = [], []
    for data in loader:
        ys.append(data.y)
        with torch.no_grad():
            out = model(data.x.to(device), data.edge_index.to(device))
        preds.append((out > 0).float().cpu())

    y, pred = torch.cat(ys, dim=0).numpy(), torch.cat(preds, dim=0).numpy()
    return f1_score(y, pred, average='micro') if pred.sum() > 0 else 0


#for epoch in range(1, 101):
#    loss = train()
#    val_f1 = test(val_loader)
#    test_f1 = test(test_loader)
#    print('Epoch: {:02d}, Loss: {:.4f}, Val: {:.4f}, Test: {:.4f}'.format(
#        epoch, loss, val_f1, test_f1))



import hyperopt

from . import register_hpo
from .advisorbase import AdvisorBaseHPOptimizer
from .suggestion.algorithm.quasi_random_search import QuasiRandomSearchAlgorithm


@register_hpo("quasi")
class QuasiAdvisorChoco(AdvisorBaseHPOptimizer):
    """
    Quasi random search algorithm in `advisor` package
    See https://github.com/tobegit3hub/advisor for the package
    See .advisorbase.AdvisorBaseHPOptimizer for more information
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method = QuasiRandomSearchAlgorithm()



import torch.nn.functional as F
from ..space import BaseSpace
from .base import BaseEstimator
from autogllight.utils.evaluation import Acc
from autogllight.utils.backend.op import *


class OneShotEstimator(BaseEstimator):
    """
    One shot estimator.

    Use model directly to get estimations.

    Parameters
    ----------
    loss_f : str
        The name of loss funciton in PyTorch
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
    """

    def __init__(self, loss_f="nll_loss", evaluation=[Acc()]):
        super().__init__(loss_f, evaluation)
        self.evaluation = evaluation

    def infer(self, model: BaseSpace, dataset, mask="train", *args, **kwargs):
        device = next(model.parameters()).device
        dset = dataset[0].to(device)
        mask = bk_mask(dset, mask)

        pred = model(dset, *args, **kwargs)[mask]
        label = bk_label(dset)
        y = label[mask]

        loss = getattr(F, self.loss_f)(pred, y)
        probs = F.softmax(pred, dim=1).detach().cpu().numpy()

        y = y.cpu()
        metrics = {
            eva.get_eval_name(): eva.evaluate(probs, y) for eva in self.evaluation
        }
        return metrics, loss



import math

import torch


def uniform(size, tensor):
    if tensor is not None:
        bound = 1.0 / math.sqrt(size)
        tensor.data.uniform_(-bound, bound)


def kaiming_uniform(tensor, fan, a):
    if tensor is not None:
        bound = math.sqrt(6 / ((1 + a ** 2) * fan))
        tensor.data.uniform_(-bound, bound)


def glorot(tensor):
    if tensor is not None:
        stdv = math.sqrt(6.0 / (tensor.size(-2) + tensor.size(-1)))
        tensor.data.uniform_(-stdv, stdv)


def glorot_orthogonal(tensor, scale):
    if tensor is not None:
        torch.nn.init.orthogonal_(tensor.data)
        scale /= (tensor.size(-2) + tensor.size(-1)) * tensor.var()
        tensor.data *= scale.sqrt()


def zeros(tensor):
    if tensor is not None:
        tensor.data.fill_(0)


def ones(tensor):
    if tensor is not None:
        tensor.data.fill_(1)


def normal(tensor, mean, std):
    if tensor is not None:
        tensor.data.normal_(mean, std)


def reset(nn):
    def _reset(item):
        if hasattr(item, "reset_parameters"):
            item.reset_parameters()

    if nn is not None:
        if hasattr(nn, "children") and len(list(nn.children())) > 0:
            for item in nn.children():
                _reset(item)
        else:
            _reset(nn)



import json
import random

from ...suggestion.models import Study
from ...suggestion.models import Trial
from ...suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm
from ...suggestion.algorithm.util import AlgorithmUtil


class RandomSearchAlgorithm(AbstractSuggestionAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    """

    def get_new_suggestions(self, study, trials=[], number=1):
        """
        Get the new suggested trials with random search.
        """

        return_trial_list = []

        # study = Study.objects.get(name=study_name)
        study_configuration_json = json.loads(study.study_configuration)
        params = study_configuration_json["params"]

        for i in range(number):
            trial = Trial.create(study.name, "RandomSearchTrial")
            parameter_values_json = {}

            for param in params:

                if param["type"] == "DOUBLE":
                    suggest_value = AlgorithmUtil.get_random_value(
                        param["minValue"], param["maxValue"]
                    )

                elif param["type"] == "INTEGER":
                    suggest_value = AlgorithmUtil.get_random_int_value(
                        param["minValue"], param["maxValue"]
                    )

                elif param["type"] == "DISCRETE":
                    feasible_point_list = [
                        float(value.strip())
                        for value in param["feasiblePoints"].split(",")
                    ]
                    suggest_value = AlgorithmUtil.get_random_item_from_list(
                        feasible_point_list
                    )

                elif param["type"] == "CATEGORICAL":
                    feasible_point_list = [
                        value.strip() for value in param["feasiblePoints"].split(",")
                    ]
                    suggest_value = AlgorithmUtil.get_random_item_from_list(
                        feasible_point_list
                    )

                parameter_values_json[param["parameterName"]] = suggest_value

            trial.parameter_values = json.dumps(parameter_values_json)
            # trial.save()
            return_trial_list.append(trial)

        return return_trial_list



import abc


class AbstractEarlyStopAlgorithm(object):

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_early_stop_trials(self, trials):
        """
        Pass the trials and return the list of trials to early stop.

        Args:
          trials: The all trials of this study.
        Returns:
          The array of trial objects.
        """
        raise NotImplementedError



"""
The implementation is based on https://github.com/tobegit3hub/advisor
"""
from __future__ import print_function
import json
import random

from suggestion.models import Study
from suggestion.models import Trial

# from abstract_algorithm import AbstractSuggestionAlgorithm
from suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm

w = 0.729844  # Inertia weight to prevent velocities becoming too large
c1 = 1.496180  # Scaling co-efficient on the social component
c2 = 1.496180  # Scaling co-efficient on the cognitive component
dimension = 20  # Size of the problem
iterations = 30
swarmSize = 30


# This class contains the code of the Particles in the swarm
class Particle:
    velocity = []
    pos = []
    pBest = []

    def __init__(self):
        for i in range(dimension):
            self.pos.append(random.random())
            self.velocity.append(0.01 * random.random())
            self.pBest.append(self.pos[i])
        return

    def updatePositions(self):
        for i in range(dimension):
            self.pos[i] = self.pos[i] + self.velocity[i]
        return

    def updateVelocities(self, gBest):
        for i in range(dimension):
            r1 = random.random()
            r2 = random.random()
            social = c1 * r1 * (gBest[i] - self.pos[i])
            cognitive = c2 * r2 * (self.pBest[i] - self.pos[i])
            self.velocity[i] = (w * self.velocity[i]) + social + cognitive
        return

    def satisfyConstraints(self):
        # This is where constraints are satisfied
        return


# This class contains the particle swarm optimization algorithm
class ParticleSwarmOptimizer:
    solution = []
    swarm = []

    def __init__(self):
        for h in range(swarmSize):
            particle = Particle()
            self.swarm.append(particle)
        return

    def optimize(self):
        for i in range(iterations):
            print("iteration ", i)
            # Get the global best particle
            gBest = self.swarm[0]
            for j in range(swarmSize):
                pBest = self.swarm[j].pBest
                if self.f(pBest) > self.f(gBest):
                    gBest = pBest
            solution = gBest
            # Update position of each paricle
            for k in range(swarmSize):
                self.swarm[k].updateVelocities(gBest)
                self.swarm[k].updatePositions()
                self.swarm[k].satisfyConstraints()
            # Update the personal best positions
            for l in range(swarmSize):
                pBest = self.swarm[l].pBest
                if self.f(self.swarm[l]) > self.f(pBest):
                    self.swarm[l].pBest = self.swarm[l].pos
        return solution

    def f(self, solution):
        # This is where the metaheuristic is defined
        return random.random()


class ParticleSwarmOptimization(AbstractSuggestionAlgorithm):
    def run_pso_demo(self):
        pso = ParticleSwarmOptimizer()
        pso.optimize()
        return

    def get_new_suggestions(self, study_name, trials=[], number=1):
        """
        Get the new suggested trials with grid search.
        """
        study = Study.objects.get(id=study_name)

        result = []
        for i in range(number):
            trial = Trial.create(study.name, "RandomSearchTrial")
            parameter_values_json = {}

            study_configuration_json = json.loads(study.study_configuration)
            params = study_configuration_json["params"]
            for param in params:
                min_value = param["minValue"]
                max_value = param["maxValue"]

                if number > 1:
                    value_step = (max_value - min_value) / (number - 1)
                else:
                    value_step = max_value - min_value
                parameter_value = min_value + value_step * i
                parameter_values_json[param["parameterName"]] = parameter_value

            trial.parameter_values = json.dumps(parameter_values_json)
            trial.save()
            result.append(trial)

        return result



from .operations import *

OPS = {
    "ZERO": lambda indim, outdim, head, dropout, concat=False: Zero(indim, outdim),
    "CONST": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="const", dropout=dropout
    ),
    "GCN": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="gcn", dropout=dropout
    ),
    "GAT": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="gat", dropout=dropout
    ),
    "SYM": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="gat_sym", dropout=dropout
    ),
    "COS": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="cos", dropout=dropout
    ),
    "LIN": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="linear", dropout=dropout
    ),
    "GENE": lambda indim, outdim, head, dropout, concat=False: GeoLayer(
        indim, outdim, head, concat, att_type="generalized_linear", dropout=dropout
    ),
}

PRIMITIVES = list(OPS.keys())



from ...suggestion.algorithm.base_chocolate_algorithm import BaseChocolateAlgorithm


class ChocolateBayesAlgorithm(BaseChocolateAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    Get the new suggested trials with bayes algorithm.
    """

    def __init__(self):
        super(ChocolateBayesAlgorithm, self).__init__("Bayes")



# "Adversarially Robust Neural Architecture Search for Graph Neural Networks"
import logging
import torch
import torch.optim
from .spos import Evolution, UniformSampler, Spos
from ..space import (
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingInputChoice,
    PathSamplingLayerChoice,
)

_logger = logging.getLogger(__name__)


class GRNA(Spos):
    """
    GRNA trainer.

    Parameters
    ----------
    n_warmup : int
        Number of epochs for training super network.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    
    Other parameters see Evolution
    """

    def __init__(
        self,
        n_warmup=1000,
        grad_clip=5.0,
        disable_progress=False,
        optimize_mode="maximize",
        population_size=100,
        sample_size=25,
        cycles=20000,
        mutation_prob=0.05,
        device="cuda",
    ):
        super().__init__(
            n_warmup,
            grad_clip,
            disable_progress,
            optimize_mode,
            population_size,
            sample_size,
            cycles,
            mutation_prob,
            device,
        )

    def _prepare(self):
        # replace choice
        self.nas_modules = []
        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)
        # to device
        self.model = self.model.to(self.device)
        self.model_optim = torch.optim.Adam(
            self.model.parameters(), lr=self.model_lr, weight_decay=self.model_wd
        )
        # controller
        self.controller = UniformSampler(self.nas_modules)

        # Evolution
        self.evolve = Evolution(
            optimize_mode="maximize",
            population_size=self.population_size,
            sample_size=self.sample_size,
            cycles=self.cycles,
            mutation_prob=self.mutation_prob,
            disable_progress=self.disable_progress,
        )

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.model, self.dataset, mask=mask)
        return metric, loss



"""
This file includes modified mutables for supporting ordered operations and also advanced choices. 
"""

from . import mutables
from .fixed import FixedArchitecture
import json
import logging
from torch import nn
from .mutables import Mutable, InputChoice, LayerChoice

_logger = logging.getLogger(__name__)


class OrderedMutable:
    """
    An abstract class with order, enabling to sort mutables with a certain rank.

    Parameters
    ----------
    order : int
        The order of the mutable
    """

    def __init__(self, order):
        self.order = order


class OrderedLayerChoice(OrderedMutable, mutables.LayerChoice):
    def __init__(
        self, order, op_candidates, reduction="sum", return_mask=False, key=None
    ):
        OrderedMutable.__init__(self, order)
        mutables.LayerChoice.__init__(self, op_candidates, reduction, return_mask, key)


class OrderedInputChoice(OrderedMutable, mutables.InputChoice):
    def __init__(
        self,
        order,
        n_candidates=None,
        choose_from=None,
        n_chosen=None,
        reduction="sum",
        return_mask=False,
        key=None,
    ):
        OrderedMutable.__init__(self, order)
        mutables.InputChoice.__init__(
            self, n_candidates, choose_from, n_chosen, reduction, return_mask, key
        )


class StrModule(nn.Module):
    """
    A shell used to wrap choices as nn.Module for non-one-shot space definition
    You can use ``map_nn`` function

    Parameters
    ----------
    name : anything
        the name of module, can be any type
    """

    def __init__(self, name):
        super().__init__()
        self.str = name

    def forward(self, *args, **kwargs):
        return self.str

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.str)


def map_nn(names):
    """
    A function used to wrap choices as nn.Module for non-one-shot space definition

    Parameters
    ----------
    name : list of anything
        the names of module, can be any type
    """
    return [StrModule(x) for x in names]


class FixedInputChoice(nn.Module):
    """
    Use to replace `InputChoice` Mutable in fix process

    Parameters
    ----------
    mask : list
        The mask indicating which input to choose
    """

    def __init__(self, mask):
        self.mask_len = len(mask)
        for i in range(self.mask_len):
            if mask[i]:
                self.selected = i
                break
        super().__init__()

    def forward(self, optional_inputs):
        if len(optional_inputs) == self.mask_len:
            return optional_inputs[self.selected]


class CleanFixedArchitecture(FixedArchitecture):
    """
    Fixed architecture mutator that always selects a certain graph, allowing deepcopy

    Parameters
    ----------
    model : nn.Module
        A mutable network.
    fixed_arc : dict
        Preloaded architecture object.
    strict : bool
        Force everything that appears in ``fixed_arc`` to be used at least once.
    verbose : bool
        Print log messages if set to True
    """

    def __init__(self, model, fixed_arc, strict=True, verbose=True):
        super().__init__(model, fixed_arc, strict, verbose)

    def replace_all_choice(self, module=None, prefix=""):
        """
        Replace all choices with selected candidates. It's done with best effort.
        In case of weighted choices or multiple choices. if some of the choices on weighted with zero, delete them.
        If single choice, replace the module with a normal module.

        Parameters
        ----------
        module : nn.Module
            Module to be processed.
        prefix : str
            Module name under global namespace.
        """

        if module is None:
            module = self.model
        for name, mutable in module.named_children():
            global_name = (prefix + "." if prefix else "") + name
            if isinstance(mutable, OrderedLayerChoice):
                chosen = self._fixed_arc[mutable.key]
                if sum(chosen) == 1 and max(chosen) == 1 and not mutable.return_mask:
                    # sum is one, max is one, there has to be an only one
                    # this is compatible with both integer arrays, boolean arrays and float arrays
                    setattr(module, name, mutable[chosen.index(1)])
                else:
                    # remove unused parameters
                    for ch, n in zip(chosen, mutable.names):
                        if ch == 0 and not isinstance(ch, float):
                            setattr(mutable, n, None)
            elif isinstance(mutable, OrderedInputChoice):
                chosen = self._fixed_arc[mutable.key]
                setattr(module, name, FixedInputChoice(chosen))
            else:
                self.replace_all_choice(mutable, global_name)


def apply_fixed_architecture(model, fixed_arc, verbose=True):
    """
    Load architecture from `fixed_arc` and apply to model.

    Parameters
    ----------
    model : torch.nn.Module
        Model with mutables.
    fixed_arc : str or dict
        Path to the JSON that stores the architecture, or dict that stores the exported architecture.
    verbose : bool
        Print log messages if set to True

    Returns
    -------
    FixedArchitecture
        Mutator that is responsible for fixes the graph.
    """

    if isinstance(fixed_arc, str):
        with open(fixed_arc) as f:
            fixed_arc = json.load(f)
    architecture = CleanFixedArchitecture(model, fixed_arc, verbose)
    architecture.reset()

    # for the convenience of parameters counting
    architecture.replace_all_choice()
    return architecture


def get_module_order(root_module):
    key2order = {}

    def apply(m):
        for name, child in m.named_children():
            if isinstance(child, Mutable):
                key2order[child.key] = child.order
            else:
                apply(child)

    apply(root_module)
    return key2order


def sort_replaced_module(k2o, modules):
    modules = sorted(modules, key=lambda x: k2o[x[0]])
    return modules


def _replace_module_with_type(root_module, init_fn, type_name, modules):
    if modules is None:
        modules = []

    def apply(m):
        for name, child in m.named_children():
            if isinstance(child, type_name):
                setattr(m, name, init_fn(child))
                modules.append((child.key, getattr(m, name)))
            else:
                apply(child)

    apply(root_module)
    return modules


def replace_layer_choice(root_module, init_fn, modules=None):
    """
    Replace layer choice modules with modules that are initiated with init_fn.

    Parameters
    ----------
    root_module : nn.Module
        Root module to traverse.
    init_fn : Callable
        Initializing function.
    modules : dict, optional
        Update the replaced modules into the dict and check duplicate if provided.

    Returns
    -------
    List[Tuple[str, nn.Module]]
        A list from layer choice keys (names) and replaced modules.
    """
    return _replace_module_with_type(root_module, init_fn, (LayerChoice,), modules)


def replace_input_choice(root_module, init_fn, modules=None):
    """
    Replace input choice modules with modules that are initiated with init_fn.

    Parameters
    ----------
    root_module : nn.Module
        Root module to traverse.
    init_fn : Callable
        Initializing function.
    modules : dict, optional
        Update the replaced modules into the dict and check duplicate if provided.

    Returns
    -------
    List[Tuple[str, nn.Module]]
        A list from layer choice keys (names) and replaced modules.
    """
    return _replace_module_with_type(root_module, init_fn, (InputChoice,), modules)



# Modified from NNI

import logging

import torch
import torch.optim
import torch.nn as nn
import torch.nn.functional as F

from .base import BaseNAS
from ..estimator.base import BaseEstimator
from ..space import BaseSpace
from ..space.nni import (
    replace_layer_choice,
    replace_input_choice,
    DartsLayerChoice,
    DartsInputChoice,
)
from tqdm import trange

_logger = logging.getLogger(__name__)


class Darts(BaseNAS):
    """
    DARTS trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    workers : int
        Workers for data loading.
    gradient_clip : float
        Gradient clipping. Set to 0 to disable. Default: 5.
    model_lr : float
        Learning rate to optimize the model.
    model_wd : float
        Weight decay to optimize the model.
    arch_lr : float
        Learning rate to optimize the architecture.
    arch_wd : float
        Weight decay to optimize the architecture.
    device : str or torch.device
        The device of the whole process
    """

    def __init__(
        self,
        num_epochs=5,
        workers=4,
        gradient_clip=5.0,
        model_lr=1e-3,
        model_wd=5e-4,
        arch_lr=3e-4,
        arch_wd=1e-3,
        device="auto",
        disable_progress=False,
    ):
        super().__init__(device=device)
        self.num_epochs = num_epochs
        self.workers = workers
        self.gradient_clip = gradient_clip
        self.model_optimizer = torch.optim.Adam
        self.arch_optimizer = torch.optim.Adam
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.arch_lr = arch_lr
        self.arch_wd = arch_wd
        self.disable_progress = disable_progress

    def search(self, space: BaseSpace, dataset, estimator):
        model_optim = self.model_optimizer(
            space.parameters(), self.model_lr, weight_decay=self.model_wd
        )

        nas_modules = []
        replace_layer_choice(space, DartsLayerChoice, nas_modules)
        replace_input_choice(space, DartsInputChoice, nas_modules)
        space = space.to(self.device)

        ctrl_params = {}
        for _, m in nas_modules:
            if m.name in ctrl_params:
                assert (
                    m.alpha.size() == ctrl_params[m.name].size()
                ), "Size of parameters with the same label should be same."
                m.alpha = ctrl_params[m.name]
            else:
                ctrl_params[m.name] = m.alpha
        arch_optim = self.arch_optimizer(
            list(ctrl_params.values()), self.arch_lr, weight_decay=self.arch_wd
        )

        with trange(self.num_epochs, disable=self.disable_progress) as bar:
            for epoch in bar:
                metric, loss = self._train_one_epoch(
                    epoch, space, dataset, estimator, model_optim, arch_optim
                )
                bar.set_postfix(loss=loss.item(), **metric)

        selection = self.export(nas_modules)
        print(selection)
        return space.parse_model(selection)

    def _train_one_epoch(
        self,
        epoch,
        model: BaseSpace,
        dataset,
        estimator,
        model_optim: torch.optim.Optimizer,
        arch_optim: torch.optim.Optimizer,
    ):
        model.train()

        # phase 1. architecture step
        arch_optim.zero_grad()
        # only no unroll here
        _, loss = self._infer(model, dataset, estimator, "val")
        loss.backward()
        arch_optim.step()

        # phase 2: child network step
        model_optim.zero_grad()
        metric, loss = self._infer(model, dataset, estimator, "train")
        loss.backward()
        # gradient clipping
        if self.gradient_clip > 0:
            nn.utils.clip_grad_norm_(model.parameters(), self.gradient_clip)
        model_optim.step()

        return metric, loss

    def _infer(self, model: BaseSpace, dataset, estimator: BaseEstimator, mask="train"):
        metric, loss = estimator.infer(model, dataset, mask=mask)
        return metric, loss

    @torch.no_grad()
    def export(self, nas_modules) -> dict:
        result = dict()
        for name, module in nas_modules:
            if name not in result:
                result[name] = module.export()
        return result



from . import BaseSpace
from .gauss_space import COAUTHOR, RandomSampler, RLSampler, Supernet


class GaussSpace(BaseSpace):
    def __init__(self, input_dim, output_dim, add_pre, args):
        super().__init__()
        self.num_layers = args.num_layers
        self.input_dim = input_dim  # data.x.size(-1)
        self.output_dim = output_dim  # num_classes, dataset.num_classes
        self.num_classes = output_dim
        self.hidden_channels = args.hidden_channels
        self.dropout = args.dropout
        self.track = args.track
        self.add_pre = add_pre
        self.args = args
        self.use_forward = True

    def build_graph(self):
        self.model = Supernet(
            self.num_layers,
            self.input_dim,
            self.output_dim,
            self.hidden_channels,
            self.dropout,
            track=self.track,
            add_pre=self.add_pre,
        ).cuda()

        if self.args.use_sampler:
            self.sampler = RLSampler(
                COAUTHOR,
                self.args.num_layers,
                epochs=self.args.epoch_sampler,
                iter=self.args.iter_sampler,
                lr=self.args.lr_sampler,
                T=self.args.T,
                entropy=self.args.entropy,
            )
        else:
            self.sampler = RandomSampler(COAUTHOR, self.args.num_layers)

    def forward(self, data, arch):
        if not self.use_forward:
            return self.prediction

        pred = self.model(data, arch)
        self.current_pred = pred
        return pred

    def keep_prediction(self):
        self.prediction = self.current_pred

    def parse_model(self, selection):
        self.use_forward = False
        return self.wrap()



# codes in this file are reproduced from https://github.com/microsoft/nni with some changes.
import copy
import torch
import torch.nn as nn
import torch.nn.functional as F
import logging

from .base import BaseNAS
from ..space import BaseSpace
from ..space import (
    BaseSpace,
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingInputChoice,
    PathSamplingLayerChoice,
    apply_fixed_architecture,
)
from tqdm import tqdm
from datetime import datetime
import numpy as np
from .rl_utils import ReinforceController, ReinforceField

LOGGER = logging.getLogger(__name__)


class Enas(BaseNAS):
    """
    ENAS trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    n_warmup : int
        Number of epochs for training super network.
    log_frequency : int
        Step count per logging.
    grad_clip : float
        Gradient clipping. Set to 0 to disable. Default: 5.
    entropy_weight : float
        Weight of sample entropy loss.
    skip_weight : float
        Weight of skip penalty loss.
    baseline_decay : float
        Decay factor of baseline. New baseline will be equal to ``baseline_decay * baseline_old + reward * (1 - baseline_decay)``.
    ctrl_lr : float
        Learning rate for RL controller.
    ctrl_steps_aggregate : int
        Number of steps that will be aggregated into one mini-batch for RL controller.
    ctrl_kwargs : dict
        Optional kwargs that will be passed to :class:`ReinforceController`.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    disable_progeress: boolean
        Control whether show the progress bar.
    device : str or torch.device
        The device of the whole process, e.g. "cuda", torch.device("cpu")
    """

    def __init__(
        self,
        num_epochs=5,
        n_warmup=100,
        log_frequency=None,
        grad_clip=5.0,
        entropy_weight=0.0001,
        skip_weight=0.8,
        baseline_decay=0.999,
        ctrl_lr=0.00035,
        ctrl_steps_aggregate=20,
        ctrl_kwargs=None,
        model_lr=5e-3,
        model_wd=5e-4,
        disable_progress=False,
        device="auto",
    ):
        super().__init__(device)
        self.num_epochs = num_epochs
        self.log_frequency = log_frequency
        self.entropy_weight = entropy_weight
        self.skip_weight = skip_weight
        self.baseline_decay = baseline_decay
        self.baseline = 0.0
        self.ctrl_steps_aggregate = ctrl_steps_aggregate
        self.grad_clip = grad_clip
        self.ctrl_kwargs = ctrl_kwargs
        self.ctrl_lr = ctrl_lr
        self.n_warmup = n_warmup
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.disable_progress = disable_progress

    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset  # .to(self.device)
        self.estimator = estimator
        # replace choice
        self.nas_modules = []

        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)

        # to device
        self.model = self.model.to(self.device)
        self.model_optim = torch.optim.Adam(
            self.model.parameters(), lr=self.model_lr, weight_decay=self.model_wd
        )
        # fields
        self.nas_fields = [
            ReinforceField(
                name,
                len(module),
                isinstance(module, PathSamplingLayerChoice) or module.n_chosen == 1,
            )
            for name, module in self.nas_modules
        ]
        self.controller = ReinforceController(
            self.nas_fields, **(self.ctrl_kwargs or {})
        )
        self.ctrl_optim = torch.optim.Adam(
            self.controller.parameters(), lr=self.ctrl_lr
        )

        # warm up supernet
        with tqdm(range(self.n_warmup), disable=self.disable_progress) as bar:
            for i in bar:
                acc, l1 = self._train_model(i)
                with torch.no_grad():
                    val_acc, val_loss = self._infer("val")
                bar.set_postfix(
                    loss=l1.item(),
                    acc=acc["acc"],
                    val_acc=val_acc["acc"],
                    val_loss=val_loss.item(),
                )

        # train
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                # try:
                acc, l1 = self._train_model(i)
                l2 = self._train_controller(i)
                """except Exception as e:
                    print(e)
                    nm = self.nas_modules
                    for i in range(len(nm)):
                        print(nm[i][1].sampled)"""
                bar.set_postfix(loss=l1.item(), acc=acc["acc"], reward_controller=l2)

        selection = self.export()
        # print(selection)
        return space.parse_model(selection)

    def _train_model(self, epoch):
        self.model.train()
        self.controller.eval()
        self.model_optim.zero_grad()
        self._resample()
        metric, loss = self._infer()
        loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip)
        self.model_optim.step()

        return metric, loss

    def _train_controller(self, epoch):
        self.model.eval()
        self.controller.train()
        self.ctrl_optim.zero_grad()
        rewards = []
        for ctrl_step in range(self.ctrl_steps_aggregate):
            self._resample()
            with torch.no_grad():
                metric, loss = self._infer(mask="val")
                reward = metric["acc"]
            rewards.append(reward)
            if self.entropy_weight:
                reward += self.entropy_weight * self.controller.sample_entropy.item()
            self.baseline = self.baseline * self.baseline_decay + reward * (
                1 - self.baseline_decay
            )
            loss = self.controller.sample_log_prob * (reward - self.baseline)
            if self.skip_weight:
                loss += self.skip_weight * self.controller.sample_skip_penalty
            loss /= self.ctrl_steps_aggregate
            loss.backward()

            if (ctrl_step + 1) % self.ctrl_steps_aggregate == 0:
                if self.grad_clip > 0:
                    nn.utils.clip_grad_norm_(
                        self.controller.parameters(), self.grad_clip
                    )
                self.ctrl_optim.step()
                self.ctrl_optim.zero_grad()

            if self.log_frequency is not None and ctrl_step % self.log_frequency == 0:
                LOGGER.info(
                    "RL Epoch [%d/%d] Step [%d/%d]  %s",
                    epoch + 1,
                    self.num_epochs,
                    ctrl_step + 1,
                    self.ctrl_steps_aggregate,
                )
        return sum(rewards) / len(rewards)

    def _resample(self):
        result = self.controller.resample()
        for name, module in self.nas_modules:
            module.sampled = result[name]

    def export(self):
        self.controller.eval()
        with torch.no_grad():
            return self.controller.resample()

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.model, self.dataset, mask=mask)
        return metric, loss



import torch
import torch.nn.functional as F
from torch.nn import Parameter
from torch_geometric.nn.inits import glorot, zeros
from torch_geometric.utils import remove_self_loops, add_self_loops, add_remaining_self_loops, softmax
from torch_scatter import scatter_add

from .message_passing import MessagePassing
#from torch_geometric.nn.conv import MessagePassing

class GeoLayer(MessagePassing):

    def __init__(self,
                 in_channels,
                 out_channels,
                 heads=1,
                 concat=True,
                 negative_slope=0.2,
                 dropout=0,
                 bias=True,
                 att_type="gat",
                 agg_type="sum",
                 pool_dim=0):
        if agg_type in ["sum", "mlp"]:
            super(GeoLayer, self).__init__('add')
        elif agg_type in ["mean", "max"]:
            super(GeoLayer, self).__init__(agg_type)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.att_type = att_type
        self.agg_type = agg_type

        # GCN weight
        #self.gcn_weight = None

        self.weight = Parameter(torch.Tensor(in_channels, heads * out_channels))
        self.att = Parameter(torch.Tensor(1, heads, 2 * out_channels))

        if bias and concat:
            self.bias = Parameter(torch.Tensor(heads * out_channels))
        elif bias and not concat:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        if self.att_type in ["generalized_linear"]:
            self.general_att_layer = torch.nn.Linear(out_channels, 1, bias=False)

        #if self.agg_type in ["mean", "max", "mlp"]:
        #    if pool_dim <= 0:
        #        pool_dim = 128
        #self.pool_dim = pool_dim
        #if pool_dim != 0:
        #    self.pool_layer = torch.nn.ModuleList()
        #    self.pool_layer.append(torch.nn.Linear(self.out_channels, self.pool_dim))
        #    self.pool_layer.append(torch.nn.Linear(self.pool_dim, self.out_channels))
        #else:
        #    pass
        self.reset_parameters()

    @staticmethod
    def norm(edge_index, num_nodes, edge_weight, improved=False, dtype=None):
        if edge_weight is None:
            edge_weight = torch.ones((edge_index.size(1), ),
                                     dtype=dtype,
                                     device=edge_index.device)

        fill_value = 1 if not improved else 2
        edge_index, edge_weight = add_remaining_self_loops(
            edge_index, edge_weight, fill_value, num_nodes)

        row, col = edge_index
        deg = scatter_add(edge_weight, row, dim=0, dim_size=num_nodes)
        deg_inv_sqrt = deg.pow(-0.5)
        deg_inv_sqrt[deg_inv_sqrt == float('inf')] = 0

        return edge_index, deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.att)
        zeros(self.bias)

        if self.att_type in ["generalized_linear"]:
            glorot(self.general_att_layer.weight)

        #if self.pool_dim != 0:
        #    for layer in self.pool_layer:
        #        glorot(layer.weight)
        #        zeros(layer.bias)

    def forward(self, x, edge_index, size=None):
        """"""
        if size is None and torch.is_tensor(x):
            edge_index, _ = remove_self_loops(edge_index)
            edge_index, _ = add_self_loops(edge_index, num_nodes=x.size(0))
        # prepare
        if torch.is_tensor(x):
            x = torch.mm(x, self.weight).view(-1, self.heads, self.out_channels)
        else:
            x = (None if x[0] is None else torch.matmul(x[0], self.weight).view(-1, self.heads, self.out_channels),
                 None if x[1] is None else torch.matmul(x[1], self.weight).view(-1, self.heads, self.out_channels))
        num_nodes = x.size(0) if torch.is_tensor(x) else size[0]
        return self.propagate(edge_index, size=size, x=x, num_nodes=num_nodes)

    def message(self, x_i, x_j, edge_index, num_nodes):

        if self.att_type == "const":
            if self.training and self.dropout > 0:
                x_j = F.dropout(x_j, p=self.dropout, training=True)
            neighbor = x_j
        #elif self.att_type == "gcn":
        #    if self.gcn_weight is None or self.gcn_weight.size(0) != x_j.size(0):  # 对于不同的图gcn_weight需要重新计算
        #        _, norm = self.norm(edge_index, num_nodes, None)
        #        self.gcn_weight = norm
        #    neighbor = self.gcn_weight.view(-1, 1, 1) * x_j
        else:
            # Compute attention coefficients.
            alpha = self.apply_attention(edge_index, num_nodes, x_i, x_j)
            alpha = softmax(alpha, edge_index[0], ptr=None, num_nodes=num_nodes)
            # Sample attention coefficients stochastically.
            if self.training and self.dropout > 0:
                alpha = F.dropout(alpha, p=self.dropout, training=True)

            neighbor = x_j * alpha.view(-1, self.heads, 1)
        #if self.pool_dim > 0:
        #    for layer in self.pool_layer:
        #        neighbor = layer(neighbor)
        return neighbor

    def apply_attention(self, edge_index, num_nodes, x_i, x_j):
        if self.att_type == "gat":
            alpha = (torch.cat([x_i, x_j], dim=-1) * self.att).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope)

        elif self.att_type == "gat_sym":
            wl = self.att[:, :, :self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels:]  # weight right
            alpha = (x_i * wl).sum(dim=-1) + (x_j * wr).sum(dim=-1)
            alpha_2 = (x_j * wl).sum(dim=-1) + (x_i * wr).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope) + F.leaky_relu(alpha_2, self.negative_slope)

        elif self.att_type == "linear":
            wl = self.att[:, :, :self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels:]  # weight right
            al = x_j * wl
            ar = x_j * wr
            alpha = al.sum(dim=-1) + ar.sum(dim=-1)
            alpha = torch.tanh(alpha)
        elif self.att_type == "cos":
            wl = self.att[:, :, :self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels:]  # weight right
            alpha = x_i * wl * x_j * wr
            alpha = alpha.sum(dim=-1)

        elif self.att_type == "generalized_linear":
            wl = self.att[:, :, :self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels:]  # weight right
            al = x_i * wl
            ar = x_j * wr
            alpha = al + ar
            alpha = torch.tanh(alpha)
            alpha = self.general_att_layer(alpha)
        else:
            raise Exception("Wrong attention type:", self.att_type)
        return alpha

    def update(self, aggr_out):
        if self.concat is True:
            aggr_out = aggr_out.view(-1, self.heads * self.out_channels)
        else:
            aggr_out = aggr_out.mean(dim=1)

        if self.bias is not None:
            aggr_out = aggr_out + self.bias
        return aggr_out

    def __repr__(self):
        return '{}({}, {}, heads={})'.format(self.__class__.__name__,
                                             self.in_channels,
                                             self.out_channels, self.heads)



"""
HPO Module for tuning hyper parameters
"""

import json
import math
import time

from tqdm import trange

from .base import BaseHPOptimizer, TimeTooLimitedError
from .suggestion.algorithm.mocmaes import MocmaesAlgorithm
from .suggestion.algorithm.random_search import RandomSearchAlgorithm
from .suggestion.models import Study


class AdvisorBaseHPOptimizer(BaseHPOptimizer):
    """
    An abstract HPOptimizer using for `advisor` package.
    See https://github.com/tobegit3hub/advisor for more information

    Attributes
    ----------
    method : .suggestion.AbstractSuggestionAlgorithm
        The algorithm class in `suggestion`
    max_evals : int
        The max rounds of evaluating HPs

    Methods
    -------
    optimize
        Optimize the HP by the method within give model and HP space
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method = None
        self.max_evals = kwargs.get("max_evals", 100)

    def _set_up(self, num_slaves, time_limit, memory_limit):
        """
        See .base.BaseHPOptimizer._set_up
        """
        self.new_trials = []
        self.xs = []
        self.best_id = None
        self.best_trainer = None
        space = self.hp_space
        current_config = self._encode_para(space)

        for i in range(num_slaves):
            self.new_trials.append(None)
            self.xs.append(None)

        study_configuration_json = {
            "goal": "MINIMIZE",
            "maxTrials": 5,
            "maxParallelTrials": 1,
            "randomInitTrials": 1,
            "params": current_config,
        }
        study_configuration = json.dumps(study_configuration_json)
        self.study = Study.create("HPO", study_configuration)
        self.config = study_configuration_json

    def _update_trials(self, pid, hp, perf):
        """
        See .base.BaseHPOptimizer._update_trials
        """
        self.new_trials[pid].status = "Completed"
        self.new_trials[pid].objective_value = perf
        if not self.best_id or perf < self.trials[self.best_id].objective_value:
            self.best_id = len(self.trials)
            self.best_hp = hp
        self.trials.append(self.new_trials[pid])

    def _get_suggestion(self, pid):
        """
        See .base.BaseHPOptimizer._get_suggestion
        """
        new_trials = self.method.get_new_suggestions(
            self.study, trials=self.trials, number=1
        )

        new_trial = new_trials[0]
        new_parameter_values_json = json.loads(new_trial.parameter_values)
        decoded_json, trial_para = self._decode_para(new_parameter_values_json)
        new_trial.parameter_values = json.dumps(trial_para)
        self.new_trials[pid] = new_trial
        self.xs[pid] = decoded_json
        return decoded_json

    def _best_hp(self):
        """
        See .base.BaseHPOptimizer._best_hp
        """
        if len(self.trials) == 0:
            return None, None
        best_perf = self.trials[self.best_id].objective_value
        decoded_json, _ = self._decode_para(
            json.loads(self.trials[self.best_id].parameter_values)
        )
        # self.logger.info("Best Parameter:")
        # self._print_info(decoded_json, best_perf)
        return self.best_trainer, best_perf

    def _setUp(self, current_config):
        study_configuration_json = {
            "goal": "MINIMIZE",
            "maxTrials": 5,
            "maxParallelTrials": 1,
            "randomInitTrials": 1,
            "params": current_config,
        }
        study_configuration = json.dumps(study_configuration_json)
        self.study = Study.create("HPO", study_configuration)
        self.config = study_configuration_json

    def optimize(self):
        """
        See .base.BaseHPOptimizer.optimize
        """
        self.trials = []
        if self.method == None:
            return

        space = self.hp_space
        current_space = self._encode_para(space)
        self._setUp(current_space)

        start_time = time.time()

        best_id = None
        best_trainer = None

        print("HPO Search Phase:\n")
        for i in trange(self.max_evals):
            if self.time_limit is not None and time.time() - start_time > self.time_limit:
                self.logger.info("Time out of limit, Epoch: {}".format(str(i)))
                break
            new_trials = self.method.get_new_suggestions(
                self.study, trials=self.trials, number=1
            )

            new_trial = new_trials[0]
            new_parameter_values_json = json.loads(new_trial.parameter_values)
            decoded_json, trial_para = self._decode_para(new_parameter_values_json)
            perf = self.f(decoded_json)
            new_trial.parameter_values = json.dumps(trial_para)
            new_trial.status = "Completed"
            new_trial.objective_value = perf
            if not best_id or perf < self.trials[best_id].objective_value:
                best_id = len(self.trials)
                best_hp = decoded_json
            else:
                del decoded_json
            self.trials.append(new_trial)
            # self._print_info(decoded_json, perf)

        if len(self.trials) == 0:
            raise TimeTooLimitedError(
                "Given time is too limited to finish one round in HPO."
            )

        best_perf = self.trials[best_id].objective_value
        # self.logger.info("Best Parameter:")

        return best_hp, best_perf



# codes in this file are reproduced from https://github.com/GraphNAS/GraphNAS with some changes.

from torch_geometric.nn import (
    GATConv,
    GCNConv,
    ChebConv,
    SAGEConv,
    GatedGraphConv,
    ARMAConv,
    SGConv,
)
import torch_geometric.nn
import torch
from torch import nn
import torch.nn.functional as F


class LinearConv(nn.Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(LinearConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.linear = torch.nn.Linear(in_channels, out_channels, bias)

    def forward(self, x, edge_index, edge_weight=None):
        return self.linear(x)

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.in_channels, self.out_channels
        )


class ZeroConv(nn.Module):
    def forward(self, x, edge_index, edge_weight=None):
        out = torch.zeros_like(x)
        out.requires_grad = True
        return out

    def __repr__(self):
        return "ZeroConv()"


class Identity(nn.Module):
    def forward(self, x, edge_index, edge_weight=None):
        return x

    def __repr__(self):
        return "Identity()"


def gnn_map(gnn_name, in_dim, out_dim, concat=False, bias=True) -> nn.Module:
    """

    :param gnn_name:
    :param in_dim:
    :param out_dim:
    :param concat: for gat, concat multi-head output or not
    :return: GNN model
    """
    if gnn_name == "gat_8":
        return GATConv(in_dim, out_dim, 8, concat=concat, bias=bias)
    elif gnn_name == "gat_6":
        return GATConv(in_dim, out_dim, 6, concat=concat, bias=bias)
    elif gnn_name == "gat_4":
        return GATConv(in_dim, out_dim, 4, concat=concat, bias=bias)
    elif gnn_name == "gat_2":
        return GATConv(in_dim, out_dim, 2, concat=concat, bias=bias)
    elif gnn_name in ["gat_1", "gat"]:
        return GATConv(in_dim, out_dim, 1, concat=concat, bias=bias)
    elif gnn_name == "gcn":
        return GCNConv(in_dim, out_dim)
    elif gnn_name == "cheb":
        return ChebConv(in_dim, out_dim, K=2, bias=bias)
    elif gnn_name == "sage":
        return SAGEConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "gated":
        return GatedGraphConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "arma":
        return ARMAConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "sg":
        return SGConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "linear":
        return LinearConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "zero":
        return ZeroConv()
    elif gnn_name == "identity":
        return Identity()
    elif hasattr(torch_geometric.nn, gnn_name):
        cls = getattr(torch_geometric.nn, gnn_name)
        assert isinstance(cls, type), "Only support modules, get %s" % (gnn_name)
        kwargs = {
            "in_channels": in_dim,
            "out_channels": out_dim,
            "concat": concat,
            "bias": bias,
        }
        kwargs = {
            key: kwargs[key]
            for key in cls.__init__.__code__.co_varnames
            if key in kwargs
        }
        return cls(**kwargs)
    raise KeyError("Cannot parse key %s" % (gnn_name))


import inspect
import sys
from torch_scatter import scatter_add
import torch_scatter

special_args = [
    "edge_index",
    "edge_index_i",
    "edge_index_j",
    "size",
    "size_i",
    "size_j",
]
__size_error_msg__ = (
    "All tensors which should get mapped to the same source "
    "or target nodes must be of same size in dimension 0."
)

is_python2 = sys.version_info[0] < 3
getargspec = inspect.getargspec if is_python2 else inspect.getfullargspec


def scatter_(name, src, index, dim_size=None):
    r"""Aggregates all values from the :attr:`src` tensor at the indices
    specified in the :attr:`index` tensor along the first dimension.
    If multiple indices reference the same location, their contributions
    are aggregated according to :attr:`name` (either :obj:`"add"`,
    :obj:`"mean"` or :obj:`"max"`).

    Args:
        name (string): The aggregation to use (:obj:`"add"`, :obj:`"mean"`,
            :obj:`"max"`).
        src (Tensor): The source tensor.
        index (LongTensor): The indices of elements to scatter.
        dim_size (int, optional): Automatically create output tensor with size
            :attr:`dim_size` in the first dimension. If set to :attr:`None`, a
            minimal sized output tensor is returned. (default: :obj:`None`)

    :rtype: :class:`Tensor`
    """

    assert name in ["add", "mean", "max"]

    op = getattr(torch_scatter, "scatter_{}".format(name))
    fill_value = -1e9 if name == "max" else 0

    out = op(src, index, 0, None, dim_size)
    if isinstance(out, tuple):
        out = out[0]

    if name == "max":
        out[out == fill_value] = 0

    return out


from torch.nn import Parameter
from torch_geometric.nn.inits import glorot, zeros
from torch_geometric.utils import (
    remove_self_loops,
    add_self_loops,
    add_remaining_self_loops,
    softmax,
)


class MessagePassing(torch.nn.Module):
    def __init__(self, aggr="add", flow="source_to_target"):
        super(MessagePassing, self).__init__()

        self.aggr = aggr
        assert self.aggr in ["add", "mean", "max"]

        self.flow = flow
        assert self.flow in ["source_to_target", "target_to_source"]

        self.__message_args__ = getargspec(self.message)[0][1:]
        self.__special_args__ = [
            (i, arg)
            for i, arg in enumerate(self.__message_args__)
            if arg in special_args
        ]
        self.__message_args__ = [
            arg for arg in self.__message_args__ if arg not in special_args
        ]
        self.__update_args__ = getargspec(self.update)[0][2:]

    def propagate(self, edge_index, size=None, **kwargs):
        r"""The initial call to start propagating messages.

        Args:
            edge_index (Tensor): The indices of a general (sparse) assignment
                matrix with shape :obj:`[N, M]` (can be directed or
                undirected).
            size (list or tuple, optional): The size :obj:`[N, M]` of the
                assignment matrix. If set to :obj:`None`, the size is tried to
                get automatically inferrred. (default: :obj:`None`)
            **kwargs: Any additional data which is needed to construct messages
                and to update node embeddings.
        """

        size = [None, None] if size is None else list(size)
        assert len(size) == 2

        i, j = (0, 1) if self.flow == "target_to_source" else (1, 0)
        ij = {"_i": i, "_j": j}

        message_args = []
        for arg in self.__message_args__:
            if arg[-2:] in ij.keys():
                tmp = kwargs.get(arg[:-2], None)
                if tmp is None:  # pragma: no cover
                    message_args.append(tmp)
                else:
                    idx = ij[arg[-2:]]
                    if isinstance(tmp, tuple) or isinstance(tmp, list):
                        assert len(tmp) == 2
                        if tmp[1 - idx] is not None:
                            if size[1 - idx] is None:
                                size[1 - idx] = tmp[1 - idx].size(0)
                            if size[1 - idx] != tmp[1 - idx].size(0):
                                raise ValueError(__size_error_msg__)
                        tmp = tmp[idx]

                    if size[idx] is None:
                        size[idx] = tmp.size(0)
                    if size[idx] != tmp.size(0):
                        raise ValueError(__size_error_msg__)

                    tmp = torch.index_select(tmp, 0, edge_index[idx])
                    message_args.append(tmp)
            else:
                message_args.append(kwargs.get(arg, None))

        size[0] = size[1] if size[0] is None else size[0]
        size[1] = size[0] if size[1] is None else size[1]

        kwargs["edge_index"] = edge_index
        kwargs["size"] = size

        for (idx, arg) in self.__special_args__:
            if arg[-2:] in ij.keys():
                message_args.insert(idx, kwargs[arg[:-2]][ij[arg[-2:]]])
            else:
                message_args.insert(idx, kwargs[arg])

        update_args = [kwargs[arg] for arg in self.__update_args__]

        out = self.message(*message_args)
        if self.aggr in ["add", "mean", "max"]:
            out = scatter_(self.aggr, out, edge_index[i], dim_size=size[i])
        else:
            pass
        out = self.update(out, *update_args)

        return out

    def message(self, x_j):  # pragma: no cover
        r"""Constructs messages in analogy to :math:`\phi_{\mathbf{\Theta}}`
        for each edge in :math:`(i,j) \in \mathcal{E}`.
        Can take any argument which was initially passed to :meth:`propagate`.
        In addition, features can be lifted to the source node :math:`i` and
        target node :math:`j` by appending :obj:`_i` or :obj:`_j` to the
        variable name, *.e.g.* :obj:`x_i` and :obj:`x_j`."""

        return x_j

    def update(self, aggr_out):  # pragma: no cover
        r"""Updates node embeddings in analogy to
        :math:`\gamma_{\mathbf{\Theta}}` for each node
        :math:`i \in \mathcal{V}`.
        Takes in the output of aggregation as first argument and any argument
        which was initially passed to :meth:`propagate`."""

        return aggr_out


class GeoLayer(MessagePassing):
    def __init__(
        self,
        in_channels,
        out_channels,
        heads=1,
        concat=True,
        negative_slope=0.2,
        dropout=0,
        bias=True,
        att_type="gat",
        agg_type="sum",
        pool_dim=0,
    ):
        if agg_type in ["sum", "mlp"]:
            super(GeoLayer, self).__init__("add")
        elif agg_type in ["mean", "max"]:
            super(GeoLayer, self).__init__(agg_type)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.att_type = att_type
        self.agg_type = agg_type

        # GCN weight
        self.gcn_weight = None

        self.weight = Parameter(torch.Tensor(in_channels, heads * out_channels))
        self.att = Parameter(torch.Tensor(1, heads, 2 * out_channels))

        if bias and concat:
            self.bias = Parameter(torch.Tensor(heads * out_channels))
        elif bias and not concat:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        if self.att_type in ["generalized_linear"]:
            self.general_att_layer = torch.nn.Linear(out_channels, 1, bias=False)

        if self.agg_type in ["mean", "max", "mlp"]:
            if pool_dim <= 0:
                pool_dim = 128
        self.pool_dim = pool_dim
        if pool_dim != 0:
            self.pool_layer = torch.nn.ModuleList()
            self.pool_layer.append(torch.nn.Linear(self.out_channels, self.pool_dim))
            self.pool_layer.append(torch.nn.Linear(self.pool_dim, self.out_channels))
        else:
            pass
        self.reset_parameters()

    @staticmethod
    def norm(edge_index, num_nodes, edge_weight, improved=False, dtype=None):
        if edge_weight is None:
            edge_weight = torch.ones(
                (edge_index.size(1),), dtype=dtype, device=edge_index.device
            )

        fill_value = 1 if not improved else 2
        edge_index, edge_weight = add_remaining_self_loops(
            edge_index, edge_weight, fill_value, num_nodes
        )

        row, col = edge_index
        deg = scatter_add(edge_weight, row, dim=0, dim_size=num_nodes)
        deg_inv_sqrt = deg.pow(-0.5)
        deg_inv_sqrt[deg_inv_sqrt == float("inf")] = 0

        return edge_index, deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.att)
        zeros(self.bias)

        if self.att_type in ["generalized_linear"]:
            glorot(self.general_att_layer.weight)

        if self.pool_dim != 0:
            for layer in self.pool_layer:
                glorot(layer.weight)
                zeros(layer.bias)

    def forward(self, x, edge_index):
        """"""
        edge_index, _ = remove_self_loops(edge_index)
        edge_index, _ = add_self_loops(edge_index, num_nodes=x.size(0))
        # prepare
        x = torch.mm(x, self.weight).view(-1, self.heads, self.out_channels)
        # x [2708,2,4] weight [1433,8]
        return self.propagate(edge_index, x=x, num_nodes=x.size(0))

    def message(self, x_i, x_j, edge_index, num_nodes):

        # x_i torch.Size([13264, 2, 4])
        # x_j torch.Size([13264, 2, 4])
        # edge_index torch.Size([2, 13264])
        # num_nodes 2708
        if self.att_type == "const":
            if self.training and self.dropout > 0:
                x_j = F.dropout(x_j, p=self.dropout, training=True)
            neighbor = x_j
        elif self.att_type == "gcn":
            if self.gcn_weight is None or self.gcn_weight.size(0) != x_j.size(
                0
            ):  # 对于不同的图gcn_weight需要重新计算
                _, norm = self.norm(edge_index, num_nodes, None)
                self.gcn_weight = norm
            neighbor = self.gcn_weight.view(-1, 1, 1) * x_j
        else:
            # Compute attention coefficients.
            alpha = self.apply_attention(edge_index, num_nodes, x_i, x_j)
            alpha = softmax(alpha, edge_index[0], num_nodes=num_nodes)
            # Sample attention coefficients stochastically.
            if self.training and self.dropout > 0:
                alpha = F.dropout(alpha, p=self.dropout, training=True)

            neighbor = x_j * alpha.view(-1, self.heads, 1)
        # pool_layer
        # (0): Linear(in_features=4, out_features=128, bias=True)
        #   (1): Linear(in_features=128, out_features=4, bias=True)
        if self.pool_dim > 0:
            # neighbor torch.Size([13264, 2, 4])
            for layer in self.pool_layer:
                neighbor = layer(neighbor)
        return neighbor

    def apply_attention(self, edge_index, num_nodes, x_i, x_j):
        if self.att_type == "gat":
            alpha = (torch.cat([x_i, x_j], dim=-1) * self.att).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope)

        elif self.att_type == "gat_sym":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            alpha = (x_i * wl).sum(dim=-1) + (x_j * wr).sum(dim=-1)
            alpha_2 = (x_j * wl).sum(dim=-1) + (x_i * wr).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope) + F.leaky_relu(
                alpha_2, self.negative_slope
            )

        elif self.att_type == "linear":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            al = x_j * wl
            ar = x_j * wr
            alpha = al.sum(dim=-1) + ar.sum(dim=-1)
            alpha = torch.tanh(alpha)
        elif self.att_type == "cos":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            alpha = x_i * wl * x_j * wr
            alpha = alpha.sum(dim=-1)

        elif self.att_type == "generalized_linear":
            wl = self.att[:, :, : self.out_channels]  # weight left
            wr = self.att[:, :, self.out_channels :]  # weight right
            al = x_i * wl
            ar = x_j * wr
            alpha = al + ar
            alpha = torch.tanh(alpha)
            alpha = self.general_att_layer(alpha)
        else:
            raise Exception("Wrong attention type:", self.att_type)
        return alpha

    def update(self, aggr_out):  # torch.Size([2708, 2, 4])
        if self.concat is True:
            aggr_out = aggr_out.view(
                -1, self.heads * self.out_channels
            )  # torch.Size([2708, 8])
        else:
            aggr_out = aggr_out.mean(dim=1)

        if self.bias is not None:
            aggr_out = aggr_out + self.bias
        return aggr_out

    def __repr__(self):
        return "{}({}, {}, heads={})".format(
            self.__class__.__name__, self.in_channels, self.out_channels, self.heads
        )

    def get_param_dict(self):
        params = {}
        key = f"{self.att_type}_{self.agg_type}_{self.in_channels}_{self.out_channels}_{self.heads}"
        weight_key = key + "_weight"
        att_key = key + "_att"
        agg_key = key + "_agg"
        bais_key = key + "_bais"

        params[weight_key] = self.weight
        params[att_key] = self.att
        params[bais_key] = self.bias
        if hasattr(self, "pool_layer"):
            params[agg_key] = self.pool_layer.state_dict()

        return params

    def load_param(self, params):
        key = f"{self.att_type}_{self.agg_type}_{self.in_channels}_{self.out_channels}_{self.heads}"
        weight_key = key + "_weight"
        att_key = key + "_att"
        agg_key = key + "_agg"
        bais_key = key + "_bais"

        if weight_key in params:
            self.weight = params[weight_key]

        if att_key in params:
            self.att = params[att_key]

        if bais_key in params:
            self.bias = params[bais_key]

        if agg_key in params and hasattr(self, "pool_layer"):
            self.pool_layer.load_state_dict(params[agg_key])



import torch
import typing as _typing


class AutoModule:
    def _initialize(self, *args, **kwargs) -> _typing.Optional[bool]:
        """ Abstract initialization method to override """
        raise NotImplementedError

    @property
    def initialized(self) -> bool:
        return self._initialized

    def initialize(self, *args, **kwargs) -> bool:
        if self._initialized:
            return self._initialized
        else:
            init_flag = self._initialize(*args, **kwargs)
            self._initialized = (
                init_flag if isinstance(init_flag, bool) else True
            )
            return self._initialized

    @property
    def device(self) -> torch.device:
        return self._device

    @device.setter
    def device(self, __device: _typing.Union[torch.device, str, int, None]):
        if type(__device) == torch.device or (
            type(__device) == str and __device.lower() != "auto"
        ) or type(__device) == int:
            self._device: torch.device = torch.device(__device)
        else:
            self._device: torch.device = torch.device(
                "cuda"
                if torch.cuda.is_available() and torch.cuda.device_count() > 0
                else "cpu"
            )

    def __init__(
            self,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        self.__hyper_parameters: _typing.Mapping[str, _typing.Any] = dict()
        self.__hyper_parameter_space: _typing.Iterable[_typing.Mapping[str, _typing.Any]] = []
        self.device = device
        self.__args: _typing.Tuple[_typing.Any, ...] = args
        self.__kwargs: _typing.Mapping[str, _typing.Any] = kwargs
        self._initialized: bool = False

    @property
    def hyper_parameters(self) -> _typing.Mapping[str, _typing.Any]:
        return self.__hyper_parameters

    @hyper_parameters.setter
    def hyper_parameters(self, hp: _typing.Mapping[str, _typing.Any]):
        self.__hyper_parameters = hp

    @property
    def hyper_parameter_space(self) -> _typing.Iterable[
        _typing.Mapping[str, _typing.Any]
    ]:
        return self.__hyper_parameter_space

    @hyper_parameter_space.setter
    def hyper_parameter_space(
            self, hp_space: _typing.Iterable[_typing.Mapping[str, _typing.Any]]
    ):
        self.__hyper_parameter_space = hp_space



from typing import Optional, Tuple
from torch_geometric.typing import Adj, OptTensor, PairTensor

import torch
from torch import Tensor
from torch.nn import Parameter
from torch_scatter import scatter_add
from torch_sparse import SparseTensor, matmul, fill_diag, sum, mul
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.utils import add_remaining_self_loops
from torch_geometric.utils.num_nodes import maybe_num_nodes

from .inits import glorot, zeros


@torch.jit._overload
def gcn_norm(
    edge_index,
    edge_weight=None,
    num_nodes=None,
    improved=False,
    add_self_loops=True,
    dtype=None,
):
    # type: (Tensor, OptTensor, Optional[int], bool, bool, Optional[int]) -> PairTensor  # noqa
    pass


@torch.jit._overload
def gcn_norm(
    edge_index,
    edge_weight=None,
    num_nodes=None,
    improved=False,
    add_self_loops=True,
    dtype=None,
):
    # type: (SparseTensor, OptTensor, Optional[int], bool, bool, Optional[int]) -> SparseTensor  # noqa
    pass


def gcn_norm(
    edge_index,
    edge_weight=None,
    num_nodes=None,
    improved=False,
    add_self_loops=True,
    dtype=None,
):

    fill_value = 2.0 if improved else 1.0

    if isinstance(edge_index, SparseTensor):
        adj_t = edge_index
        if not adj_t.has_value():
            adj_t = adj_t.fill_value(1.0, dtype=dtype)
        if add_self_loops:
            adj_t = fill_diag(adj_t, fill_value)
        deg = sum(adj_t, dim=1)
        deg_inv_sqrt = deg.pow_(-0.5)
        deg_inv_sqrt.masked_fill_(deg_inv_sqrt == float("inf"), 0.0)
        adj_t = mul(adj_t, deg_inv_sqrt.view(-1, 1))
        adj_t = mul(adj_t, deg_inv_sqrt.view(1, -1))
        return adj_t

    else:
        num_nodes = maybe_num_nodes(edge_index, num_nodes)

        if edge_weight is None:
            edge_weight = torch.ones(
                (edge_index.size(1),), dtype=dtype, device=edge_index.device
            )

        if add_self_loops:
            edge_index, tmp_edge_weight = add_remaining_self_loops(
                edge_index, edge_weight, fill_value, num_nodes
            )
            assert tmp_edge_weight is not None
            edge_weight = tmp_edge_weight

        row, col = edge_index[0], edge_index[1]
        deg = scatter_add(edge_weight, col, dim=0, dim_size=num_nodes)
        deg_inv_sqrt = deg.pow_(-0.5)
        deg_inv_sqrt.masked_fill_(deg_inv_sqrt == float("inf"), 0)
        return edge_index, deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]


class GCNConv(MessagePassing):
    r"""The graph convolutional operator from the `"Semi-supervised
    Classification with Graph Convolutional Networks"
    <https://arxiv.org/abs/1609.02907>`_ paper

    .. math::
        \mathbf{X}^{\prime} = \mathbf{\hat{D}}^{-1/2} \mathbf{\hat{A}}
        \mathbf{\hat{D}}^{-1/2} \mathbf{X} \mathbf{\Theta},

    where :math:`\mathbf{\hat{A}} = \mathbf{A} + \mathbf{I}` denotes the
    adjacency matrix with inserted self-loops and
    :math:`\hat{D}_{ii} = \sum_{j=0} \hat{A}_{ij}` its diagonal degree matrix.
    The adjacency matrix can include other values than :obj:`1` representing
    edge weights via the optional :obj:`edge_weight` tensor.

    Its node-wise formulation is given by:

    .. math::
        \mathbf{x}^{\prime}_i = \mathbf{\Theta} \sum_{j}
        \frac{1}{\sqrt{\hat{d}_j \hat{d}_i}} \mathbf{x}_j

    with :math:`\hat{d}_i = 1 + \sum_{j \in \mathcal{N}(i)} e_{j,i}`, where
    :math:`e_{j,i}` denotes the edge weight from source node :obj:`i` to target
    node :obj:`j` (default: :obj:`1`)

    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        improved (bool, optional): If set to :obj:`True`, the layer computes
            :math:`\mathbf{\hat{A}}` as :math:`\mathbf{A} + 2\mathbf{I}`.
            (default: :obj:`False`)
        cached (bool, optional): If set to :obj:`True`, the layer will cache
            the computation of :math:`\mathbf{\hat{D}}^{-1/2} \mathbf{\hat{A}}
            \mathbf{\hat{D}}^{-1/2}` on first execution, and will use the
            cached version for further executions.
            This parameter should only be set to :obj:`True` in transductive
            learning scenarios. (default: :obj:`False`)
        add_self_loops (bool, optional): If set to :obj:`False`, will not add
            self-loops to the input graph. (default: :obj:`True`)
        normalize (bool, optional): Whether to add self-loops and apply
            symmetric normalization. (default: :obj:`True`)
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.
    """

    _cached_edge_index: Optional[Tuple[Tensor, Tensor]]
    _cached_adj_t: Optional[SparseTensor]

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        improved: bool = False,
        cached: bool = False,
        add_self_loops: bool = True,
        normalize: bool = True,
        bias: bool = True,
        **kwargs
    ):

        kwargs.setdefault("aggr", "add")
        super(GCNConv, self).__init__(**kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.improved = improved
        self.cached = cached
        self.add_self_loops = add_self_loops
        self.normalize = normalize

        self._cached_edge_index = None
        self._cached_adj_t = None

        self.weight = Parameter(torch.Tensor(in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        zeros(self.bias)
        self._cached_edge_index = None
        self._cached_adj_t = None

    def forward(
        self, x: Tensor, edge_index: Adj, edge_weight: OptTensor = None
    ) -> Tensor:
        """"""

        if self.normalize:
            if isinstance(edge_index, Tensor):
                cache = self._cached_edge_index
                if cache is None:
                    edge_index, edge_weight = gcn_norm(  # yapf: disable
                        edge_index,
                        edge_weight,
                        x.size(self.node_dim),
                        self.improved,
                        self.add_self_loops,
                        dtype=x.dtype,
                    )
                    if self.cached:
                        self._cached_edge_index = (edge_index, edge_weight)
                else:
                    edge_index, edge_weight = cache[0], cache[1]

            elif isinstance(edge_index, SparseTensor):
                cache = self._cached_adj_t
                if cache is None:
                    edge_index = gcn_norm(  # yapf: disable
                        edge_index,
                        edge_weight,
                        x.size(self.node_dim),
                        self.improved,
                        self.add_self_loops,
                        dtype=x.dtype,
                    )
                    if self.cached:
                        self._cached_adj_t = edge_index
                else:
                    edge_index = cache

        x = torch.matmul(x, self.weight)

        # propagate_type: (x: Tensor, edge_weight: OptTensor)
        out = self.propagate(edge_index, x=x, edge_weight=edge_weight, size=None)

        if self.bias is not None:
            out += self.bias

        return out

    def message(self, x_j: Tensor, edge_weight: OptTensor) -> Tensor:
        if edge_weight is None:
            return x_j
        else:
            return edge_weight.view(-1, 1) * x_j

    def message_and_aggregate(self, adj_t: SparseTensor, x: Tensor) -> Tensor:
        return matmul(adj_t, x, reduce=self.aggr)

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.in_channels, self.out_channels
        )



import torch
import torch_geometric.nn as pygnn
import torch.nn.functional as F
from .module.conv import Conv

COAUTHOR = [
    'gatv2',
    'gcn',
    'sage',
    'linear',
    'gin',
    'graph'
]

class Supernet(torch.nn.Module):
    def __init__(self, n_layers, n_input, n_output, n_hidden, dropout, space=COAUTHOR, arch=None, track=True, add_pre=False):
        super().__init__()
        self.n_layers = n_layers
        self.n_hidden = n_hidden
        self.dropout = dropout

        self.add_pre = add_pre

        if add_pre:
            self.preprocess = torch.nn.Sequential(
                torch.nn.Linear(n_input, n_hidden),
                torch.nn.BatchNorm1d(n_hidden, track_running_stats=track),
                torch.nn.ReLU(),
                torch.nn.Dropout(dropout)
            )

        if arch is None:
            self.space = [space] * n_layers
        else:
            self.space = [[a] for a in arch]

        self.convs = torch.nn.ModuleList()

        for i in range(n_layers):
            in_feat = n_input if (i == 0 and not add_pre) else n_hidden
            out_feat = n_hidden if i < n_layers - 1 else n_output
            if i < n_layers - 1:
                self.convs.append(Conv(self.space[i], in_feat, out_feat, pos=2, act=F.relu, dropout=dropout, bn=True, res=False, track=track))
            else:
                self.convs.append(Conv(self.space[i], in_feat, out_feat, pos=0))

    def forward(self, data, arch):
        x, edge_index = data.x, data.edge_index
        if self.add_pre:
            x = self.preprocess(x)
        for i, (conv, a) in enumerate(zip(self.convs, arch)):
            x = conv(x, edge_index, a)
        return x.log_softmax(dim=-1)

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
        if self.add_pre:
            self.preprocess[0].reset_parameters()
            self.preprocess[1].reset_parameters()

if __name__ == '__main__':
    from itertools import product
    archs = [list(a) for a in product(*[COAUTHOR for _ in range(2)])]
    import pickle
    pickle.dump(archs, open('models/coauthor/archs.pkl', 'wb'))



import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import BaseNAS
from ..space import BaseSpace
from ..space.nni import (
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingLayerChoice,
    PathSamplingInputChoice,
)
from tqdm import tqdm
import numpy as np
import logging as _logging

nas_logger: _logging.Logger = _logging.getLogger("NAS")


class RandomSearch(BaseNAS):
    """
    Uniformly random architecture search

    Parameters
    ----------
    device : str or torch.device
        The device of the whole process, e.g. "cuda", torch.device("cpu")
    num_epochs : int
        Number of epochs planned for training.
    disable_progeress: boolean
        Control whether show the progress bar.
    """

    def __init__(
        self,
        device="auto",
        num_epochs=400,
        disable_progress=False,
        hardware_metric_limit=None,
        select_metric="acc",
    ):
        super().__init__(device)
        self.num_epochs = num_epochs
        self.disable_progress = disable_progress
        self.hardware_metric_limit = hardware_metric_limit
        self.select_metric = select_metric

    def init_search(self):
        self.nas_modules = []
        k2o = get_module_order(self.space)
        replace_layer_choice(self.space, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.space, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)
        selection_range = {}
        for k, v in self.nas_modules:
            selection_range[k] = len(v)
        self.selection_dict = selection_range
        space_size = np.prod(list(selection_range.values()))

        nas_logger.info(f"Selection range {selection_range}")
        nas_logger.info(f"Space size {space_size}")

    def search(self, space: BaseSpace, dset, estimator):
        self.estimator = estimator
        self.dataset = dset
        self.space = space

        self.init_search()

        arch_perfs = []
        cache = {}
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                selection = self.sample()
                vec = tuple(list(selection.values()))
                if vec not in cache:
                    self.arch = space.parse_model(selection)
                    metric, loss = self._infer(mask="val")
                    metric = metric[self.select_metric]
                    arch_perfs.append([metric, selection])
                    cache[vec] = metric
                bar.set_postfix(acc=metric, max_acc=max(cache.values()))
        selection = arch_perfs[np.argmax([x[0] for x in arch_perfs])][1]
        print(selection)
        arch = space.parse_model(selection)
        return arch

    def sample(self):
        # uniformly sample
        selection = {}
        for k, v in self.selection_dict.items():
            selection[k] = np.random.choice(range(v))
        return selection

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.arch, self.dataset, mask=mask)
        return metric, loss



import hyperopt

from . import register_hpo
from .advisorbase import AdvisorBaseHPOptimizer
from .suggestion.algorithm.chocolate_random_search import ChocolateRandomSearchAlgorithm

# @register_hpo("randomchoco")


class RandAdvisorChoco(AdvisorBaseHPOptimizer):
    def __init__(self, args):
        super().__init__(args)
        self.method = ChocolateRandomSearchAlgorithm()
